from pyspatial.globalmaptiles import GlobalMercator
from pyspatial.raster import TiledWebRaster

zoom = 13
twr = TiledWebRaster("/tmp/95000_45000_rgb/", zoom)


def test_point():
    zoom = 13
    lat, lng = 38.76479194327964,-89.23713684082031  # lat, lng of Carlyle Lake in MO
    # The color should be blue
    # Testing that pixel coords match
    GM = GlobalMercator()
    mx, my = GM.LatLonToMeters(lat, lng)
    mx, my = GM.LatLonToMeters(lat, lng)
    assert twr._to_pixels(mx, my) == GM.MetersToPixels(mx, my, zoom)



"""
Copyright (c) 2016, Granular, Inc.
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import pickle
import pyspatial.vector as vt
from pyspatial.utils import projection_from_string, ALBERS_N_AMERICA
from pyspatial.utils import projection_from_epsg
from pyspatial.spatiallib import haversine
from osgeo import ogr
from nose.tools import assert_raises, assert_almost_equal

base = os.path.abspath(os.path.dirname(__file__))
get_path = lambda x: os.path.join(base, "data/vector", x)

rect_str = """
{
        "type": "Polygon",
        "coordinates": [
          [
            [
              -122.58956909179688,
              37.73379707124429
            ],
            [
              -122.58956909179688,
              37.80218877920469
            ],
            [
              -122.4587631225586,
              37.80218877920469
            ],
            [
              -122.4587631225586,
              37.73379707124429
            ],
            [
              -122.58956909179688,
              37.73379707124429
            ]
          ]
        ]
      }"""
rect = ogr.CreateGeometryFromJson(rect_str)

farallon_str = """
{
    "type": "Polygon",
        "coordinates": [
          [
            [
              -123.04550170898436,
              37.6816466602918
            ],
            [
              -123.04550170898436,
              37.72293542866175
            ],
            [
              -122.97409057617188,
              37.72293542866175
            ],
            [
              -122.97409057617188,
              37.6816466602918
            ],
            [
              -123.04550170898436,
              37.6816466602918
            ]
          ]
        ]
}"""

farallon = ogr.CreateGeometryFromJson(farallon_str)


def __test_read():
    path = ("http://www2.census.gov/geo/tiger/GENZ2014/shp/"
            "cb_2014_us_state_500k.zip")

    assert len(vt.read_layer(path)[0]) == 56

    path = get_path("cb_2014_us_state_500k.zip")
    assert len(vt.read_layer(path)[0]) == 56

    paths = ["http://eric.clst.org/wupl/Stuff/gz_2010_us_040_00_500k.json",
             get_path("gz_2010_us_040_00_500k.json")]

    for path in paths:
        assert len(vt.read_geojson(path)[0]) == 52

    geojson_str = open(paths[1]).read()
    assert len(vt.read_geojson(geojson_str)[0]) == 52


class TestVectorLayer:
    @classmethod
    def setup_class(cls):
        path1 = get_path("clu/four_shapes_2il_2ca.geojson")
        path2 = get_path("gz_2010_us_040_00_500k.json")
        path3 = get_path("bay_area_counties.geojson")
        path4 = get_path("bay_area_zips.geojson")

        cls.vl1, cls.df1 = vt.read_geojson(path1)
        cls.vl2, cls.df2 = vt.read_geojson(path2)
        cls.counties, cls.df3 = vt.read_geojson(path3, index="NAME")
        cls.sf = "San Francisco"
        proj = projection_from_epsg()
        rect.AssignSpatialReference(proj)
        farallon.AssignSpatialReference(proj)
        cls.counties[cls.sf] = cls.counties[cls.sf].Difference(farallon)
        cls.zips, cls.df4 = vt.read_geojson(path4, index="ZCTA5CE10")
        p = get_path("clu/four_shapes_2il_2ca.p")
        cls.df = pickle.load(open(p))
        assert isinstance(cls.counties, vt.VectorLayer)
        assert isinstance(cls.counties["San Francisco"], ogr.Geometry)

    def test_read(self):
        states, df = vt.read_layer(get_path("cb_2014_us_state_500k.zip"))
        assert isinstance(states, vt.VectorLayer)
        assert len(states) == 56
        co, co_df = vt.read_layer(get_path("clu/clu_public_a_co095.shp"))
        assert isinstance(co, vt.VectorLayer)

    def test_from_series(self):
        series = self.df["__geometry__"]
        assert isinstance(vt.from_series(series), vt.VectorLayer)

    def test_predicates(self):
        sf = self.counties["San Francisco"]
        assert isinstance(sf, ogr.Geometry)
        sf_ids = self.zips.within(sf, index_only=True)
        assert all(map(lambda x: x.startswith("941"), sf_ids))

    def test_intersection(self):
        sf = self.counties["San Francisco"]
        vl = self.zips.intersection(sf)
        assert abs(vl.unary_union().area - vt.to_shapely(sf).area) < 1e-3

    def test_read_index(self):
        path = get_path("cb_2014_us_state_500k.zip")
        vl, df = vt.read_layer(path, index="STUSPS")
        assert all([a == b for a, b in zip(vl.index, df.index)])
        vl, df = vt.read_layer(path, index=xrange(5, 61))
        assert_raises(ValueError, vt.read_layer, path, 0, xrange(5, 56))

    def test_ipredicates(self):
        path = get_path("cb_2014_us_state_500k.zip")
        vl, df = vt.read_layer(path, index="STUSPS")
        clu_path = get_path("clu/clu_public_a_il189.shp")
        clus, clusdf = vt.read_layer(clu_path)
        assert vl.iintersects(clus[0])[0] == "IL"
        assert len(vl.iwithin(clus[0])) == 0
        assert "IL" not in vl.idisjoint(clus[0])
        centroid = vl["CA"].Centroid()
        centroid.AssignSpatialReference(vl.proj)
        # California contains its centroid
        assert vl.icontains(centroid)[0] == "CA"
        # Clus are within IL
        assert len(clus.head().iwithin(vl["IL"])) > 0
        # No states within clus
        assert len(vl.iwithin(clus.head()[0])) == 0
        assert len(vl[["CA"]].boundingboxes().iwithin(vl["CA"])) == 0

    def test_distance(self):
        clu_path = get_path("clu/clu_public_a_il189.shp")
        clus, clusdf = vt.read_layer(clu_path)
        path = get_path("cb_2014_us_state_500k.zip")
        vl, vldf = vt.read_layer(path, index="STUSPS")
        s = clus[:5].distances(vl["RI"], proj="albers")/1.e3
        assert s.count() == 5
        assert s.min() > 1440 and s.max() < 1441
        d = vl.boundingboxes().distances(vl["MI"], proj='albers')/1.e3
        assert abs(d["CA"] - 1853.3812112445789) < 1e-6
        # Compute the haversine (great-circle) distance to MI
        MI = vl["MI"].Centroid().GetPoints()[0]
        d = (vl.to_wgs84().centroids(format="Series")
             .map(lambda x: haversine(x, MI))/1.e3)
        assert abs(d["WA"] - 2706.922595) < 1e-6

    def test_to_json(self):
        with open(get_path("RI.json")) as inf:
            exp = inf.read()
        act, _ = vt.read_layer(get_path("cb_2014_us_state_20m.zip"),
                               index="STUSPS")
        assert exp == act[["RI"]].to_json()

    def test_set_theoretic(self):
        proj = projection_from_string(ALBERS_N_AMERICA)
        counties = self.counties.transform(proj)
        sf = [self.sf]

        union_exp = 183.026345584
        union_act = counties[sf].union(rect)[self.sf]
        assert_almost_equal(union_act.GetArea()/1e6, union_exp)

        intersection_exp = 31.0072128793
        intersection_act = counties[sf].intersection(rect)[self.sf]

        assert_almost_equal(intersection_act.GetArea()/1e6, intersection_exp)

        symdifference_exp = 152.019132704
        symdifference_act = counties[sf].difference(rect, kind="symmetric")
        assert_almost_equal(symdifference_act[self.sf].GetArea()/1e6,
                            symdifference_exp)

        ldifference_exp = 95.5399654593
        ldifference_act = counties[sf].difference(rect, kind="left")
        assert_almost_equal(ldifference_act[self.sf].GetArea()/1e6,
                            ldifference_exp)

        rdifference_exp = 56.4791672452
        rdifference_act = counties[sf].difference(rect, kind="right")
        assert_almost_equal(rdifference_act[self.sf].GetArea()/1e6,
                            rdifference_exp)


def test_intersects():
    shape, _ = vt.read_geojson(get_path('test_shape.json'))

    with open(get_path('test_soils.json')) as f:
        soils = f.read()
        soils = vt.read_geojson(soils)[0]

    soils.intersects(shape[0])
    assert (shape[0].IsValid())


def test_to_json():
    shape, _ = vt.read_geojson(get_path('test_shape.json'))
    # This shape requires more precision when serializing
    j = shape.to_json(precision=15)
    jj, _ = vt.read_geojson(j)
    assert jj[0].IsValid()



"""
Copyright (c) 2016, Granular, Inc. 
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met: 

  * Redistributions of source code must retain the above copyright notice, this list of conditions 
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
    following disclaimer in the documentation and/or other materials provided with the distribution. 
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products 
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import numpy as np

from pyspatial.raster import rasterize, read_catalog
from pyspatial.raster import RasterBand
from pyspatial.vector import read_layer


cwd = os.getcwd()
base = os.path.abspath(os.path.dirname(__file__))


def get_path(x):
    return os.path.join(base, "data", x)


class TestRasterDatasetTiled:
    @classmethod
    def setup_class(cls):
        os.chdir(base)
        # Let's add some CLUs for Washington County, IL
        cls.vl, _ = read_layer(get_path("vector/clu/clu_public_a_il189.shp"),
                               index="uniqueid")

        # Figure out which raster files we'll need.
        cls.dataset = read_catalog(get_path("../catalog/cdl_2014.json"))

        # Create a RasterBand for the raster. Raster data is stored
        # and read from there.
        cls.rb = RasterBand(get_path("raster/95000_45000.tif"))

        # Convert CLUs to pixels.
        cls.px_shps = cls.dataset.to_pixels(cls.vl)

    @classmethod
    def teardown_class(cls):
        os.chdir(cwd)

    def test_single_shape_should_have_equivalent_counts(self):

        # Read in raster projection (applies to all raster tiles).
        rast_proj = self.dataset.proj

        # Compute counts for a single shape using single-tile computation for reference.
        shp = self.px_shps[0]
        mask = rasterize(shp, ext_outline=0, int_outline=1).T
        minx, miny, maxx, maxy = shp.bounds
        pts = (np.argwhere(mask>0) + np.array([minx, miny])).astype(int)
        counts_reference = np.bincount(self.rb[pts[:,1],pts[:, 0]], minlength=256)

        # Now, compute counts for a single shape using RasterDataset, and compare against single-tile computation.
        values = np.array(self.dataset.get_values_for_pixels(pts)).astype(np.uint8)
        counts = np.bincount(values, minlength=256)
        assert(np.array_equal(counts, counts_reference))

    # Compare all shapes between single-tile computation and RasterDataset computation.
    def test_all_shapes_should_have_equivalent_counts(self):

        # Read in raster projection (applies to all raster tiles).
        rast_proj = self.dataset.proj

        # All shapes in the layer should already be in pixel coords. Check that we have the right number of them.
        assert (len(self.px_shps) == 23403)

        # Compute counts for a single shape using single-tile computation for reference.
        for i, shp in enumerate(self.px_shps):
            # Rasterize the shape, and find list of all points. This is the same for either case.
            mask = rasterize(shp, ext_outline=0, int_outline=1).T
            minx, miny, maxx, maxy = shp.bounds
            pts = (np.argwhere(mask>0) + np.array([minx, miny])).astype(int)

            # Tiled data (in RasterDataset): Compute counts.
            values_td = np.array(self.dataset.get_values_for_pixels(pts)).astype(np.uint8)
            counts_td = np.bincount(values_td, minlength=256)

            # Single-tile Dataset: Compute counts for reference, and compare for equality.
            counts_ref = np.bincount(self.rb[pts[:,1],pts[:, 0]], minlength=256)
            assert(np.array_equal(counts_td, counts_ref))



"""
Copyright (c) 2016, Granular, Inc. 
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met: 

  * Redistributions of source code must retain the above copyright notice, this list of conditions 
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
    following disclaimer in the documentation and/or other materials provided with the distribution. 
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products 
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
import pyspatial.raster as rst
import pyspatial.vector as vt
from pyspatial import fileutils
import unittest
import boto

# check if google cloud storage config exists
gs_access_key_id_exists = boto.config.get('Credentials', 'gs_access_key_id') is not None
gs_secret_access_key_exists = boto.config.get('Credentials', 'gs_secret_access_key') is not None

gs_config_exists = gs_access_key_id_exists and gs_secret_access_key_exists

# get the bucket name from env variables if it exists
gs_bucket_name = os.getenv('PS_GS_DEFAULT_BUCKET', 'pyspatial')

base = os.path.abspath(os.path.dirname(__file__))

def delete_gs_key(path):
    storage_uri = boto.storage_uri(gs_bucket_name, 'gs')
    bucket = storage_uri.get_bucket(gs_bucket_name)
    key = bucket.lookup(path)
    key.delete()

@unittest.skipIf(not gs_config_exists, "requires google cloud storage configuration")
def test_raster_io_gs():
    filepath = 'data/raster/prism.tif'
    gs_path = 'gs://%s/%s' % (gs_bucket_name, filepath)

    # write a raster file to gs
    outf = fileutils.open(gs_path, "wb")
    with open(os.path.join(base, filepath)) as inf:
        outf.write(inf.read())
    outf.close()

    # read the raster file from gs
    rd = rst.read_raster(gs_path)
    assert isinstance(rd, rst.RasterDataset)

    # clean up
    delete_gs_key(filepath)

@unittest.skipIf(not gs_config_exists, "requires google cloud storage configuration")
def test_vector_json_io_gs():
    filepath = 'data/vector/test_shape.json'
    gs_path = 'gs://%s/%s' % (gs_bucket_name, filepath)

    # write a vector json file to gs
    shape, _ = vt.read_geojson(os.path.join(base, filepath))
    shape.to_json(path=gs_path, precision=15)

    # read the vector json file from gs
    shape_json = vt.fetch_geojson(gs_path)
    geojson, _ = vt.read_geojson(shape_json)

    # ensure it is valid
    assert geojson[0].IsValid()

    # clean up
    delete_gs_key(filepath)

@unittest.skipIf(not gs_config_exists, "requires google cloud storage configuration")
def test_vector_shp_io_gs():
    filepath = 'data/vector/clu/clu_public_a_co095.shp'
    gs_path = 'gs://%s/%s' % (gs_bucket_name, filepath)

    # write a vector shape file to gs
    co, co_df = vt.read_layer(os.path.join(base, filepath))
    co.to_shapefile(gs_path, df=co_df)

    # read the vector shape file from gs
    #co, co_df = vt.read_layer(gs_path)
    #assert isinstance(co, vt.VectorLayer)

    # clean up
    delete_gs_key(filepath)


import ctypes
cimport cython
import numpy as np
cimport numpy as np
from libc.stdint cimport uintptr_t
from libc.math cimport sin, cos, exp, sqrt, floor
cdef extern from "math.h":
    double sin(double theta)
    double cos(double theta)
    double tan(double theta)
    double acos(double x)
    double asin(double x)
    double floor(double x)
    int floor(double x)

import math

cdef inline double pi(): return math.pi
cdef inline double radians(double x): return PI*x/180.
cdef inline double degrees(double x): return 180.*x/PI
cdef PI = pi()

from shapely.coords import required

DTYPE = np.uint8
ctypedef np.uint8_t DTYPE_t


cdef inline float to_pixel(float a, float A, float a_px_size):
    return (a - A)/a_px_size

@cython.boundscheck(False)
def create_image_array(np.ndarray[DTYPE_t, ndim=2] rast,
                       np.ndarray[DTYPE_t, ndim=2] colors):

    cdef:
        unsigned int xsize = rast.shape[0]
        unsigned int ysize = rast.shape[1]
        unsigned int i, j, c
        np.ndarray[DTYPE_t, ndim=3] img = np.zeros([xsize, ysize, 4], dtype=DTYPE)

    for i in xrange(xsize):
        for j in xrange(ysize):
            img[i, j, :] = colors[rast[i,j]]
    return img

def to_pixels(float lon, float lat, float minLon, float maxLat,
              float lon_px_size, float lat_px_size):
    lon_px = to_pixel(lon, minLon, lon_px_size)
    lat_px = to_pixel(lat, maxLat, lat_px_size)
    return lon_px, lat_px

def grid_for_pixel(int grid_size, np.int_t x, np.int_t y):
    x_grid = x - x % grid_size
    y_grid = y - y % grid_size
    return (x_grid, y_grid)


def sub(tup, float minx, float miny):
    return (tup[0] - minx, tup[1] - miny)


@cython.boundscheck(False)
cpdef adjust_coords(geom, float minx, float miny):
    ob = required(geom)
    array = ob.__array_interface__

    cdef:
         unsigned int xsize = array['shape'][0]
         np.ndarray[np.float64_t, ndim=2] res = np.zeros([xsize, 2], dtype=np.float64)
         unsigned int sm, sn, n, dx

    n = 2
    dx = 0
    if array.get('strides', None):
        sm = array['strides'][0]/sizeof(dx)
        sn = array['strides'][1]/sizeof(dx)
    else:
        sm = n
        sn = 1

    # Make pointer to the coordinate array
    if isinstance(array['data'], ctypes.Array):
        cp = <double *><uintptr_t>ctypes.addressof(array['data'])
    else:
        cp = <double *><uintptr_t>array['data'][0]

    #print xsize, array["shape"][1]

    for i in xrange(xsize):
        res[i, 0] = cp[sm*i] - minx
        res[i, 1] = cp[sm*i+sn] - miny

    return res

cdef double K0 = 0.9996
cdef double E = 0.00669438
cdef double E2 = E * E
cdef double E3 = E2 * E
cdef double E_P2 = E / (1.0 - E)
cdef double SQRT_E = sqrt(1 - E)
cdef double _E = (1 - SQRT_E) / (1 + SQRT_E)
cdef double _E2 = _E * _E
cdef double _E3 = _E2 * _E
cdef double _E4 = _E3 * _E
cdef double _E5 = _E4  * _E
cdef double M1 = (1 - E / 4 - 3 * E2 / 64 - 5 * E3 / 256)
cdef double M2 = (3 * E / 8 + 3 * E2 / 32 + 45 * E3 / 1024)
cdef double M3 = (15 * E2 / 256 + 45 * E3 / 1024)
cdef double M4 = (35 * E3 / 3072)
cdef double P2 = (3. / 2 * _E - 27. / 32 * _E3 + 269. / 512 * _E5)
cdef double P3 = (21. / 16 * _E2 - 55. / 32 * _E4)
cdef double P4 = (151. / 96 * _E3 - 417. / 128 * _E5)
cdef double P5 = (1097. / 512 * _E4)
cdef double R = 6378137


cdef int latlon_to_zone_number(double latitude, double longitude):
    if 56 <= latitude < 64 and 3 <= longitude < 12:
        return 32

    if 72 <= latitude <= 84 and longitude >= 0:
        if longitude <= 9:
            return 31
        elif longitude <= 21:
            return 33
        elif longitude <= 33:
            return 35
        elif longitude <= 42:
            return 37

    return floor((longitude + 180.) / 6.) + 1

cdef double zone_number_to_central_longitude(int zone_number):
    return (zone_number - 1) * 6 - 180 + 3


cdef struct LatLon:
    double lat
    double lon


@cython.overflowcheck(False)
@cython.nonecheck(False)
@cython.cdivision(True)
cdef LatLon from_latlon(double latitude, double longitude):

    lat_rad = radians(latitude)
    lat_sin = sin(lat_rad)
    lat_cos = cos(lat_rad)

    lat_tan = lat_sin / lat_cos
    lat_tan2 = lat_tan * lat_tan
    lat_tan4 = lat_tan2 * lat_tan2

    zone_number = latlon_to_zone_number(latitude, longitude)

    lon_rad = radians(longitude)
    central_lon = zone_number_to_central_longitude(zone_number)
    central_lon_rad = radians(central_lon)

    n = R / sqrt(1 - E * lat_sin**2)
    c = E_P2 * lat_cos**2

    a = lat_cos * (lon_rad - central_lon_rad)
    a2 = a * a
    a3 = a2 * a
    a4 = a3 * a
    a5 = a4 * a
    a6 = a5 * a

    m = R * (M1 * lat_rad -
             M2 * sin(2 * lat_rad) +
             M3 * sin(4 * lat_rad) -
             M4 * sin(6 * lat_rad))

    easting = K0 * n * (a +
                        a3 / 6 * (1 - lat_tan2 + c) +
                        a5 / 120 * (5 - 18 * lat_tan2 + lat_tan4 + 72 * c - 58 * E_P2)) + 500000

    northing = K0 * (m + n * lat_tan * (a2 / 2 +
                                        a4 / 24 * (5 - lat_tan2 + 9 * c + 4 * c**2) +
                                        a6 / 720 * (61 - 58 * lat_tan2 + lat_tan4 + 600 * c - 330 * E_P2)))

    if latitude < 0:
        northing += 10000000

    return LatLon(easting, northing)

# Note, had to elevation because shapely can give coordinates
# in 2 or 3 dimensions when using ops.transform
def to_utm(np.float64_t lon, np.float64_t lat, ele=None):
    """Compute the coordinates in UTM, will ignore elevation"""
    latlon = from_latlon(lat, lon)
    return latlon.lon, latlon.lat

def haversine(tuple coord1, tuple coord2):
    """Given two (lng, lat) tuples, returns the distance between them in
    meters."""
    cdef double lat1
    cdef double lng1
    cdef double lat2
    cdef double lng2
    lng1, lat1 = coord1
    lng2, lat2 = coord2

    if lat1 > 90 or lat1 < -90 or lat2 > 90 or lat2 < -90:
        raise ValueError("Invalid latitude (should be between +/- 90)")
    if lng1 > 180 or lng1 < -180 or lng2 > 180 or lng2 < -180:
        raise ValueError("Invalid longitude (should be between +/- 180)")

    cdef double ph1
    cdef double ph2
    cdef double theta1
    cdef double theta2
    cdef double c
    cdef double arc

    phi1 = (90.0 - lat1) * 0.0174532925
    phi2 = (90.0 - lat2) * 0.0174532925
    theta1 = lng1 * 0.0174532925
    theta2 = lng2 * 0.0174532925

    c = (sin(phi1) * sin(phi2) * cos(theta1 - theta2) + cos(phi1) * cos(phi2))
    arc = acos(c)
    return arc * 6367444.7


"""
Copyright (c) 2016, Granular, Inc. 
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met: 

  * Redistributions of source code must retain the above copyright notice, this list of conditions 
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
    following disclaimer in the documentation and/or other materials provided with the distribution. 
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products 
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os

# Scipy stack
import numpy as np
import pandas as pd

# Spatial
from pyspatial.vector import read_layer, read_geojson
from pyspatial.raster import read_catalog

from nose.tools import timed

cwd = os.getcwd()
base = os.path.abspath(os.path.dirname(__file__))


def get_path(x):
    return os.path.join(base, "data", x)


def compute_stats(values, weights):
    # values = np.array(row, dtype=np.uint8)
    counts = np.bincount(values, weights=weights, minlength=256)
    zeros = np.zeros(counts.shape)
    total = 1. * np.sum(counts)
    return counts/total if total > 0 else zeros


class TestRasterQuery:
    @classmethod
    def setup_class(cls):
        os.chdir(base)

        # Let's add some CLUs for Washington County, IL
        cls.vl, _ = read_layer(get_path("vector/clu/clu_public_a_il189.shp"),
                               index="uniqueid")
        cls.one_shape_id = "3850036893056913813r"
        cls.vl_one_shape = cls.vl[[cls.one_shape_id]]
        cls.areas = cls.vl.areas()
        cls.areas = cls.areas/cls.areas.sum()

        # Read expected output computed via single-tile computation
        # from a file, so we can compare against that.
        cls.df_expected = pd.read_csv(get_path("expected.csv"),
                                      index_col=0)
        cls.df_expected.rename(columns=lambda x: int(x), inplace=True)
        cls.df_expected.applymap(lambda x: int(x))

        # For single shape output, we only want to compare our data against that shape.
        cls.df_expected_one_shape = cls.df_expected.loc[cls.one_shape_id]

    @classmethod
    def teardown_class(cls):
        os.chdir(cwd)

    def make_dataframe(self, generator):
        index = []
        res = []

        for r in generator:
            res.append(compute_stats(r.values, r.weights))
            index.append(r.id)

        return pd.DataFrame(res, index=index)

    def test_shapes_outside_raster_should_be_filtered(self):
        # 2 shapes in IL that should be in range,
        # plus 2 shapes in CA that should be out of range.
        p = get_path("vector/clu/four_shapes_2il_2ca.geojson")
        vl_outside, vl_df = read_geojson(p)
        assert (len(vl_outside.keys()) == 4)

        dataset_catalog_file = get_path("../catalog/cdl_2014.json")
        rd = read_catalog(dataset_catalog_file)
        df = self.make_dataframe(rd.query(vl_outside))
        assert (len(df.index) == 4)
        sums = df.sum(axis=1).map(int)
        assert df[sums < 1e-6].shape[0] == 2
        assert df[(sums - 1).map(np.abs) < 1e-6].shape[0] == 2

    # Compute term frequency for cdl_2014 on a tiled dataset for all shapes,
    # and compare against our saved single-tile computation.
    @timed(60)
    def test_term_frequency_tiled_all_shapes(self):
        dataset_catalog_file = get_path("../catalog/cdl_2014.json")
        rd = read_catalog(dataset_catalog_file)
        df = self.make_dataframe(rd.query(self.vl))

        assert (len(df.index) == 23403)

        # Compare against expected output computed via single-tile computation
        # read from a file.
        df_distance = (self.df_expected - df).applymap(np.abs)
        corn_error = (df_distance[1]*self.areas).sum()
        soy_error = (df_distance[5]*self.areas).sum()
        assert(corn_error < 0.02), corn_error
        assert(soy_error < 0.02), soy_error

    # Removed this functionality as it needs more thought
    # Compute term frequency for cdl_2014 on a tiled dataset for all shapes,
    # and compare against our saved single-tile computation.
    #@timed(45)
    #def test_term_frequency_tiled_all_shapes_with_index(self):
    #    dataset_catalog_file = get_path("../catalog/cdl_2014_with_index.json")
    #    rd = read_catalog(dataset_catalog_file)
    #    df = self.make_dataframe(rd.query(self.vl))

    #    assert (len(df.index) == 23403)

        # Compare against expected output computed via single-tile computation
        # read from a file.
    #    df_distance = (self.df_expected - df).applymap(np.abs)
    #    corn_error = (df_distance[1]*self.areas).sum()
    #    soy_error = (df_distance[5]*self.areas).sum()
    #    assert(corn_error < 0.02), corn_error
    #    assert(soy_error < 0.02), soy_error

    # Compute term frequency for cdl_2014 on an untiled dataset for all shapes,
    # and compare against our saved single-tile computation.
    @timed(60)
    def test_term_frequency_untiled_all_shapes(self):
        dataset_catalog_file = get_path("../catalog/cdl_2014_untiled.json")
        rd = read_catalog(dataset_catalog_file)
        df = self.make_dataframe(rd.query(self.vl))

        assert (len(df.index) == 23403)

        # Compare against expected output computed via single-tile computation
        # read from a file.
        df_distance = (self.df_expected - df).applymap(np.abs)
        corn_error = (df_distance[1]*self.areas).sum()
        soy_error = (df_distance[5]*self.areas).sum()
        assert(corn_error < 0.02), corn_error
        assert(soy_error < 0.02), soy_error

    # Compute term frequency for cdl_2014 on a tiled dataset for one shape,
    # and compare against our saved single-tile computation.
    @timed(0.5)
    def test_term_frequency_tiled_one_shape(self):
        dataset_catalog_file = get_path("../catalog/cdl_2014.json")
        rd = read_catalog(dataset_catalog_file)
        df = self.make_dataframe(rd.query(self.vl_one_shape))

        # Compare against expected output computed via single-tile computation
        # read from a file.
        df_distance = (self.df_expected_one_shape - df.loc[self.one_shape_id])\
            .apply(np.abs)
        assert(df_distance[1] < 0.05)
        assert(df_distance[5] < 0.05)

    # Compute term frequency for cdl_2014 on an untiled dataset for one shape,
    # and compare against our saved single-tile computation.
    @timed(0.5)
    def test_term_frequency_untiled_one_shape(self):
        dataset_catalog_file = get_path("../catalog/cdl_2014_untiled.json")
        rd = read_catalog(dataset_catalog_file)
        df = self.make_dataframe(rd.query(self.vl_one_shape))
        # Compare against expected output computed via single-tile computation
        # read from a file.
        df_distance = (self.df_expected_one_shape - df.loc[self.one_shape_id])\
            .apply(np.abs)

        assert(df_distance[1] < 0.05)
        assert(df_distance[5] < 0.05)

    @timed(5)
    def test_co_ne_border(self):
        p = get_path("vector/clu/clu_public_a_co095.shp")
        vl, _ = read_layer(p)
        p = get_path("vector/co_ne_border.geojson")
        co_ne_border, df = read_geojson(p)
        vl = vl.within(co_ne_border.bbox())

        rd = read_catalog(get_path("../catalog/co_soil.json"))
        for r in rd.query(vl):
            r
            #compute_stats(r.values, r.weights)

        rd = read_catalog(get_path("../catalog/co_soil_bad.json"))
        failed = False
        try:
            for r in rd.query(vl):
                r
                #compute_stats(r.values, r.weights)
        except IndexError:
            failed = True

        assert failed

    # Test if tilepaths were defined from a different working directory
    # than the python code
    def test_unconventional_tilepath(self):
        dataset_catalog_file = get_path("../catalog/cdl_2014_tilepath.json")
        rd = read_catalog(dataset_catalog_file, workdir='data/raster')
        df = self.make_dataframe(rd.query(self.vl_one_shape))
        assert (len(df.index) == 1)



"""
Copyright (c) 2016, Granular, Inc.
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import json
import os
import math
import re
from uuid import uuid4
from six import string_types
#Scipy
import numpy as np
from skimage.transform import downscale_local_mean

#Geo
from osgeo import gdal, osr
from osgeo.gdalconst import GA_ReadOnly
from osgeo.osr import SpatialReference
from shapely import wkb, ops
from shapely.affinity import scale
from shapely.geometry import box
from skimage.io import imsave

from PIL import Image, ImageDraw
from pyspatial import fileutils

from pyspatial import spatiallib as slib
from pyspatial.vector import read_geojson, to_geometry, bounding_box
from pyspatial.vector import VectorLayer
from pyspatial.utils import projection_from_epsg
from pyspatial import globalmaptiles

NP2GDAL_CONVERSION = {
    "uint8": 1,
    "uint16": 2,
    "int16": 3,
    "uint32": 4,
    "int32": 5,
    "float32": 6,
    "float64": 7,
    "complex64": 10,
    "complex128": 11,
}

GDAL2NP_CONVERSION = {v: k for k, v in NP2GDAL_CONVERSION.items()}
TILE_REGEX = re.compile('([0-9]+)_([0-9]+)\.tif')


def rasterize(shp, ext_outline=False, ext_fill=True, int_outline=False,
              int_fill=False, scale_factor=4):

    """Convert a vector shape to a raster. Assumes the shape has already
    been transformed in to a pixel based coordinate system. The algorithm
    checks for the intersection of each point in the shape with
    a pixel grid created by the bounds of the shape. Partial overlaps
    are estimated by scaling the image in X and Y by the scale factor,
    rasterizing the shape, and downscaling (using mean), back to the
    bounds of the original shape.

    Parameters
    ----------

    shp: shapely.Polygon or Multipolygon
        The shape to rasterize

    ext_outline: boolean (default False)
        Include the outline of the shape in the raster

    ext_fill: boolean (default True)
        Fill the shape in the raster

    int_outline: booelan (default False)
        Include the outline of the interior shapes

    int_fill: boolean (default False):
        Fill the interior shapes

    scale_factor: int (default 4)
        The amount to scale the shape in X, Y before downscaling. The
        higher this number, the more precise the estimate of the overlap.


    Returns
    -------
    np.ndarray representing the rasterized shape.
    """
    sf = int(scale_factor)

    minx, miny, maxx, maxy = map(int, shp.bounds)
    if minx == maxx and miny == maxy:
        return np.array([[1.]])

    elif maxy > miny and minx == maxx:
        n = maxy - miny + 1
        return np.zeros([n, 1]) + 1./n

    elif maxy == miny and minx < maxx:
        n = maxx - minx + 1
        return np.zeros([1, n]) + 1./n

    if ((maxx - minx + 1) + (maxy - miny + 1)) <= 2*sf:
        sf = 1

    shp = scale(shp, xfact=sf, yfact=sf)
    minx, miny, maxx, maxy = shp.bounds
    width = int(maxx - minx + 1)
    height = int(maxy - miny + 1)

    img = Image.new('L', (width, height), 0)
    _shp = shp.geoms if hasattr(shp, "geoms") else [shp]

    ext_outline = int(ext_outline)
    ext_fill = int(ext_fill)
    int_outline = int(int_outline)
    int_fill = int(int_fill)

    for pg in _shp:
        ext_pg = [(x-minx, y-miny) for x, y in pg.exterior.coords]
        ImageDraw.Draw(img).polygon(ext_pg, outline=ext_outline, fill=ext_fill)
        for s in pg.interiors:
            int_pg = [(x-minx, y-miny) for x, y in s.coords]
            ImageDraw.Draw(img).polygon(int_pg, outline=int_outline,
                                        fill=int_fill)

    return downscale_local_mean(np.array(img), (sf, sf))


class RasterBase(object):
    """
    Provides methods and attributes common to both RasterBand and
    RasterDataset, particularly for converting shapes to pixels
    in the raster coordinate space. Stores a coordinate system for a raster.

    Parameters
    ----------
    RasterXSize, RasterYSize: int
        Number of pixels in the width and height respectively.

    geo_transform : list of float
        GDAL coefficients for GeoTransform (defines boundaries and pixel size
        for a raster in lat/lon space).

    proj: osr.SpatialReference
        The spatial projection for the raster.

    Attributes
    ----------
    xsize, ysize: int
        Number of pixels in the width and height respectively.

    geo_transform : list of float
        GDAL coefficients for GeoTransform (defines boundaries and pixel size
        for a raster in lat/lon space).

    min_lon: float
         The minimum longitude in proj coordinates

    min_lat: float
         The minimum latitude in proj coordinates

    max_lat: float
         The maximum latitude in proj coordinates

    lon_px_size: float
         Horizontal size of the pixel

    lat_px_size: float
         Vertical size of the pixel

    proj: osr.SpatialReference
         The spatial projection for the raster.
    """
    def __init__(self, RasterXSize, RasterYSize, geo_transform, proj):
        self.geo_transform = geo_transform
        self.xsize = RasterXSize
        self.ysize = RasterYSize
        self.RasterXSize = self.xsize
        self.RasterYSize = self.ysize
        self.min_lon = self.geo_transform[0]
        self.max_lat = self.geo_transform[3]
        self.min_lat = self.geo_transform[3] + self.geo_transform[5]*self.ysize
        self.lon_px_size = abs(self.geo_transform[1])
        self.lat_px_size = self.geo_transform[5]
        self.pixel_area = abs(self.lon_px_size * self.lat_px_size)
        self.proj = proj

    def _to_pixels(self, lon, lat, alt=None):
        """Convert a point from lon/lat to pixel coordinates.  Note,
        the altitude is currently ignored.

        Parameters
        ----------
        lon: float
            Longitude of point

        lat: float
            Latitude of point

        Returns
        -------
        list of int
            (longitude in pixel space, latitude in pixel space).
            Rounded to the nearest pixel.
        """
        lon_px, lat_px = slib.to_pixels(lon, lat, self.min_lon,
                                        self.max_lat, self.lon_px_size,
                                        self.lat_px_size)
        return int(lon_px), int(lat_px)

    def shape_to_pixel(self, geom):
        """Takes a feature and returns a shapely object transformed into the
        pixel coords.

        Parameters
        ----------
        feat : osgeo.ogr.Geometry
            Feature to be transformed.

        Returns
        -------
        shapely.Polygon
            Feature in pixel coordinates.
        """
        shp = wkb.loads(geom.ExportToWkb())
        return ops.transform(self._to_pixels, shp)

    def to_pixels(self, vector_layer):
        """Takes a vector layer and returns list of shapely geometry
        transformed in pixel coordinates. If the projection of the
        vector_layer is different than the raster band projection, it
        transforms the coordinates first to raster projection.

        Parameters
        ----------
        vector_layer : VectorLayer
            Shapes to be transformed.

        Returns
        -------
        list of shapely.Polygon
            Shapes in pixel coordinates.
        """
        if self.proj.ExportToProj4() != vector_layer.proj.ExportToProj4():
            vector_layer = vector_layer.transform(self.proj)
        return [self.shape_to_pixel(geom) for geom in vector_layer]

    def to_raster_coord(self, pxx, pxy):
        """Convert pixel corrdinates -> raster coordinates"""
        if not (0 <= pxx < self.RasterXSize):
            raise ValueError("Invalid x coordinate: %s" % pxx)

        if not (0 <= pxy < self.RasterYSize):
            raise ValueError("Invalid x coordinate: %s" % pxx)

        # urx, ury are the upper right coordinates
        # xsize, ysize, are the pixel sizes
        urx, xsize, _, ury, _, ysize = self.geo_transform
        return (urx + pxx * xsize, ury + ysize * pxy)

    def to_geometry_grid(self,  minx, miny, maxx, maxy):
        """Convert pixels into a geometry grid. All values should be in
        pixel cooridnates.

        Returns
        -------
        VectorLayer with index a tuple of the upper left corner coordinate
        of each pixel.
        """

        xs = np.arange(minx, maxx+1)
        ys = np.arange(miny, maxy+1)

        x, y = np.meshgrid(xs, ys)
        index = []
        boxes = []
        for i in range(x.shape[0]):
            for j in range(x.shape[1]):
                x1, y1 = self.to_raster_coord(x[i, j], y[i, j])
                x2, y2 = self.to_raster_coord(x[i, j] + 1, y[i, j] + 1)
                boxes.append(bounding_box((x1, x2, y1, y2), self.proj))
                index.append((int(x[i, j]), int(y[i, j])))

        return VectorLayer(boxes, index=index, proj=self.proj)

    def GetGeoTransform(self):
        """Returns affine transform from GDAL for describing the relationship
        between raster positions (in pixel/line coordinates) and georeferenced
        coordinates.

        Returns
        -------
        min_lon: float
             The minimum longitude in raster coordinates.

        lon_px_size: float
             Horizontal size of each pixel.

        geo_transform[2] : float
            Not used in our case. In general, this would be used if the
            coordinate system had rotation or shearing.

        max_lat: float
             The maximum latitude in raster coordinates.

        lat_px_size: float
             Vertical size of the pixel.

        geo_transform[5] : float
            Not used in our case. In general, this would be used if the
            coordinate system had rotation or shearing.

        References
        ----------
        http://www.gdal.org/gdal_tutorial.html
        """
        return self.geo_transform

    def get_extent(self):
        """Returns extent in raster coordinates.

        Returns
        -------
        xmin : float
            Minimum x-value (lon) of extent in raster coordinates.

        xmax : float
            Maximum x-value (lon) of extent in raster coordinates.

        ymin : float
            Minimum y-value (lat) of extent in raster coordinates.

        ymax : float
            Maximum y-value (lat) of extent in raster coordinates.
        """
        ymax = self.max_lat
        xmin = self.min_lon
        ymin = ymax + self.lat_px_size*self.ysize
        xmax = xmin + self.lon_px_size*self.xsize
        return (xmin, xmax, ymin, ymax)

    def bbox(self):
        """Returns bounding box of raster in raster coordinates.

        Returns
        -------
        ogr.Geometry
            Bounding box in raster coordinates:
                (xmin : float
                    minimum longitude (leftmost)

                 ymin : float
                    minimum latitude (bottom)

                 xmax : float
                    maximum longitude (rightmost)

                 ymax : float
                    maximum latitude (top))
        """
        (xmin, xmax, ymin, ymax) = self.get_extent()
        return to_geometry(box(xmin, ymin, xmax, ymax),
                           proj=self.proj)


class RasterBand(RasterBase, np.ndarray):
    def __new__(cls, ds, band_number=1):
        """
        Create an in-memory representation for a single band in
        a raster. (0,0) in pixel coordinates represents the
        upper left corner of the raster which corresponds to
        (min_lon, max_lat).  Inherits from ndarray, so you can
        use it like a numpy array.

        Parameters
        ----------
        ds: gdal.Dataset

        band_number: int
            The band number to use

        Attributes
        ----------
        data: np.ndarray[xsize, ysize]
             The raster data
        """

        if not isinstance(ds, gdal.Dataset):
            path = fileutils.get_path(ds)
            ds = gdal.Open(path, GA_ReadOnly)

        band = ds.GetRasterBand(band_number)
        if band is None:
            msg = "Unable to load band %d " % band_number
            msg += "in raster %s" % ds.GetDescription()
            raise ValueError(msg)

        gdal_type = band.DataType
        dtype = np.dtype(GDAL2NP_CONVERSION[gdal_type])
        self = np.asarray(band.ReadAsArray().astype(dtype)).view(cls)

        self.gdal_type = gdal_type
        proj = SpatialReference()
        proj.ImportFromWkt(ds.GetProjection())
        geo_transform = ds.GetGeoTransform()

        # Initialize the base class with coordinate information.
        RasterBase.__init__(self, ds.RasterXSize, ds.RasterYSize,
                            geo_transform, proj)

        self.nan = band.GetNoDataValue()

        #self = np.ma.masked_equal(self, band.GetNoDataValue(), copy=False)
        ctable = band.GetColorTable()
        if ctable is not None:
            self.colors = np.array([ctable.GetColorEntry(i)
                                    for i in range(256)],
                                   dtype=np.uint8)
        else:
            self.colors = None

        ds = None
        return self

    def __init__(self, ds, band_number=1):
        pass

    def __array_finalize__(self, obj):
        # see InfoArray.__array_finalize__ for comments
        if obj is None:
            return

        cols = ['xsize', 'ysize', 'geo_transform',
                'RasterXSize', 'RasterYSize',
                'min_lon', 'min_lat', 'max_lat',
                'lon_px_size', 'lat_px_size',
                'pixel_area', 'proj', 'gdal_type',
                'colors', 'nan']

        for c in cols:
            setattr(self, c, getattr(obj, c, None))

    def to_gdal(self, driver="MEM", path=''):
        """Convert to a gdal dataset."""
        drv = gdal.GetDriverByName(driver)
        ds = drv.Create(path, self.xsize, self.ysize, 1, self.gdal_type)
        ds.SetGeoTransform(self.GetGeoTransform())
        ds.SetProjection(self.proj.ExportToWkt())
        band = ds.GetRasterBand(1)
        if self.colors is not None:
            ctable = gdal.ColorTable()
            for i, c in enumerate(self.colors):
                ctable.SetColorEntry(i, tuple(c))
                band.SetColorTable(ctable)
        band.WriteArray(self)
        band.FlushCache()
        return ds

    def transform(self, proj, size=None, method="nneighbour"):
        """
        A sample function to reproject and resample a GDAL dataset from within
        Python. The idea here is to reproject from one system to another, as well
        as to change the pixel size. The procedure is slightly long-winded, but
        goes like this:

        1. Set up the two Spatial Reference systems.
        2. Open the original dataset, and get the geotransform
        3. Calculate bounds of new geotransform by projecting the UL corners
        4. Calculate the number of pixels with the new projection & spacing
        5. Create an in-memory raster dataset
        6. Perform the projection
        """
        methods = {"mean": gdal.GRA_Average,
                   "bilinear": gdal.GRA_Bilinear,
                   "cubic": gdal.GRA_Cubic,
                   "cubic-spline": gdal.GRA_CubicSpline,
                   "lanczos": gdal.GRA_Lanczos,
                   "mode": gdal.GRA_Mode,
                   "nneighbour": gdal.GRA_NearestNeighbour}

        if method not in methods:
            raise ValueError("methods must be one of: %s" % methods.keys())

        tx = osr.CoordinateTransformation(self.proj, proj)

        geo_t = self.GetGeoTransform()
        if size is None:
            x_size = self.RasterXSize
            y_size = self.RasterYSize
        else:
            x_size, y_size = size

        # Work out the boundaries of the new dataset in the target projection
        (ulx, uly, ulz) = tx.TransformPoint(geo_t[0], geo_t[3])
        (lrx, lry, lrz) = tx.TransformPoint(geo_t[0] + geo_t[1]*x_size,
                                            geo_t[3] + geo_t[5]*y_size)
        x_px_size = (lrx - ulx) * 1./x_size
        y_px_size = (lry - uly) * 1./y_size

        mem_drv = gdal.GetDriverByName('MEM')
        src = self.to_gdal()

        dest = mem_drv.Create('', x_size, y_size, 1, self.gdal_type)
        # Calculate the new geotransform
        new_geo = (ulx, x_px_size, geo_t[2], uly, geo_t[4], y_px_size)
        # Set the geotransform
        dest.SetGeoTransform(new_geo)
        dest.SetProjection(proj.ExportToWkt())

        # Perform the projection/resampling
        res = gdal.ReprojectImage(src, dest, self.proj.ExportToWkt(),
                                  proj.ExportToWkt(),
                                  methods[method])
        assert res == 0
        rb = RasterBand(dest)
        rb.colors = np.copy(self.colors)
        dest = None
        src = None
        return rb

    def to_wgs84(self, method="nneighbour"):
        return self.transform(projection_from_epsg(4326), method=method)

    def save(self, path, format="GTiff"):
        self.to_gdal(format, path)

    def to_rgb(self):
        return slib.create_image_array(self, self.colors)

    def save_png(self, path):
        imsave(path, self.to_rgb())


class RasterQueryResult:
    """
    Container class to hold the result of a raster query.

    Attributes
    ----------
    id : str or int
        The id of the shape in the vector layer

    coordinates : np.ndarray
        The requested raster pixel coordinates

    values: np.ndarray
        The values of the intersected pixels in the raster

    weights: np.ndarray
        The fraction of the polygon intersecting with the pixel
    """
    def __init__(self, id, coordinates, values, weights):
        self.id = id
        self.coordinates = coordinates
        self.values = values
        self.weights = weights


class RasterDataset(RasterBase):
    """
    Raster representation that supports tiled and untiled datasets, and
    allows querying with a set of shapes.

    Raster may be tiled or untiled. A RasterDataset
    object may be queried one or multiple times with a set of shapes (in a
    VectorLayer). We also try to match the attribute and method names of
    gdal.Dataset when possible to allow for easy porting of caller code
    to use this class instead of gdal.Dataset, as this class transparently
    works on an untiled gdal.Dataset, in addition to the added functionality
    to handle tiled datasets.

    Parameters
    ----------
    path_or_ds: str or gdal.Dataset

    xsize: int
         The number of pixels in the X coordinate

    ysize: int
         The number of pixels in the Y coordinate

    geo_transform: tuple

    proj: osr.SpatialReference
        The spa
    grid_size: int (default=None)
        Number of pixels for each tile. Assumes that each tile is square.

    index: dict (default=None)
        Dictionary matching the Geojson spec describing boundary of each file.
        Typically generated by gdaltindex.

    tile_structure: str (default="%d_%d.tif")
        A string describing the file structure and the format of the tiles.
        In case of use of gdal2tiles tiles, this must be set to "%d/%d.png".

    Attributes
    ----------
    path : str
        Path to raster data files.

    grid_size : int
        Number of pixels in width or height of each grid tile. If set to None,
        that indicates this is an untiled raster.

    raster_bands : RasterBand, or
                    dict of (list of int): RasterBand
        Dictionary storing raster arrays that have been read from disk.
        If untiled, this is set at initialization to the whole raster. If
        tiled, these are read in lazily as needed. Index is (x_grid, y_grid)
        where x_grid is x coordinate of leftmost pixel in this tile relative
        to minLon (and is a multiple of grid_size), and y_grid is y coordinate
        of uppermost pixel in this tile relative to maxLat (and is also a
        multiple of grid_size). See notes below for more information on how
        data is represented here.

    shapes_in_tiles : dict of (int, int): set of str
        What shapes are left to be processed in each tile. Key is (minx, maxy)
        of tile (upper left corner), and value is set of ids of shapes. This
        is initially set in query(), and shape ids are removed from this data
        structure for a tile once they have been processed. Tiles can be
        cleared from memory when there are no shapes left in their set.

    Notes
    -----
    Raster representation (tiled and untiled):
    The core functionality of this class is to look up pixel values (for a
    shape or set of shapes) in the raster. To do this, we store the raster in
    a 2D-array of pixels relative to (0,0) being the upper left corner aka
    (min_lon, max_lat) in lon/lat coordinates. We can then convert vector
    shapes into pixel space, and look up their values directly in the raster
    array. For an untiled raster, we can read in the raster directly during
    initialization.

    Tiled representation:
    For a tiled dataset (ie. the data is split into multiple files), we
    still treat the overall raster upper left corner as (0,0), and
    recognize that each tile has a position relative to the overall raster
    pixel array. We store each tile in a 2D array in a dictionary keyed by
    the tile position relative to the overall raster position in pixel space.
    For example, a pixel at (118, 243) in a tiled dataset with grid size = 100
    would be stored in raster_bands[(100, 200)][18][43]. As a memory
    utilization and performance enhancement, we lazily read tiles from disk
    when they are first needed and store them in raster_bands{} (for the
    lifetime of the RasterDataset object). If memory turns out to be a
    problem, it might make sense to store these in a LRU cache instead.

    TODOs
    -----

    * Added band number
    * Add support for color tables and raster attributes
    """

    def __init__(self, path_or_ds, xsize, ysize, geo_transform, proj,
                 tile_regex=TILE_REGEX, grid_size=None, index=None,
                 tile_structure=None, tms_z=None):

        ds = None
        self.band_count = None

        if not isinstance(path_or_ds, gdal.Dataset):
            path = path_or_ds
        else:
            ds = path_or_ds
            path = ds.GetDescription()
            self.band_count = ds.RasterCount

        self.path = path
        self.proj = proj
        self.grid_size = grid_size
        self.index = index

        if tile_structure:
            self.tile_structure = tile_structure
        else:
            self.tile_structure = "%d_%d.tif"

        self.raster_bands = {}
        self.shapes_in_tiles = {}
        self.tile_regex = tile_regex
        self.index = index
        self.grid_size = grid_size
        self.dtype = None

        # Initialize the base class with coordinate information.
        RasterBase.__init__(self, xsize, ysize, geo_transform, proj)

        # Read raster file now if this is an untiled data set.
        if self.grid_size is None:
            if ds is None:
                self.raster_bands = read_vsimem(self.path)
            else:
                if self.band_count == 1:
                    self.raster_bands = RasterBand(ds)
                    self.dtype = self.raster_bands.dtype
                else:
                    self.raster_bands = [RasterBand(ds, band_number=(i+1))
                                         for i in range(self.band_count)]
                    self.dtype = [r.dtype for r in self.raster_bands]

        if self.tile_structure:
            self.tile_regex = self.tile_structure.replace('%d','([0-9]+)').replace('.','\.')
            self.tile_regex = re.compile(self.tile_regex)

        ds = None
        path_or_ds = None

    def _get_value_for_pixel(self, px):
        """Look up value for a pixel in raster space.

        Parameters
        ----------
        px : np.array
            Pixel coordinates for 1 point: [x_coord, y_coord]

        Returns
        -------
        dtype
            Value in raster at pixel coordinates specified by px.
            Type is determined by GDAL2NP_CONVERSION from RasterBand data
            type.

        """
        x_grid, y_grid = self._get_grid_for_pixel(px)

        # Look up the value in the x,y offset in the grid tile we just found
        # or read, and return it.
        x_px = px[0] - x_grid
        y_px = px[1] - y_grid

        # If we haven't already read this grid tile into memory, do so now,
        # and store it in raster_bands for future queries to access.
        if (x_grid, y_grid) not in self.raster_bands:
            filename = self.path + self.tile_structure % (x_grid, y_grid)
            self.raster_bands[(x_grid, y_grid)] = read_vsimem(filename)
            if self.dtype is None:
                self.dtype = self.raster_bands[(x_grid, y_grid)].dtype

        # Look up the grid tile for this pixel.
        raster = self.raster_bands[(x_grid, y_grid)]

        return raster[y_px][x_px]

    def _get_grid_for_pixel(self, px):
        """Compute the min_x, min_y of the tile that contains pixel,
        which can also be used for looking up the tile in raster_bands.

        Parameters
        ----------
        px : np.array
            Pixel coordinates for 1 point: [x_coord, y_coord]

        Returns
        -------
        list of int
            (min_x, min_y) of tile that contains px, in pixel coordinates.

        """
        return slib.grid_for_pixel(self.grid_size, px[0], px[1])

    def get_values_for_pixels(self, pxs):
        """Look up values for a list of pixels in raster space.

        Parameters
        ----------
        pxs : np.array
            Array of pixel coordinates. Each row is [x_coord, y_coord]
            for one point.

        Returns
        -------
        list of dtype
            List of values in raster at pixel coordinates specified in pxs.
            Type is determined by GDAL2NP_CONVERSION from RasterBand data
            type.
        """
        # Untiled case: Use the 1-file raster array we read in at
        # initialization.
        if self.grid_size is None:
            if isinstance(self.raster_bands, list):
                return [r[pxs[:, 1], pxs[:, 0]] for r in self.raster_bands]
            else:
                return self.raster_bands[pxs[:, 1], pxs[:, 0]]
        # Tiled case: Compute the grid tile to read, and the x,y offset in
        # that tile.
        else:
            return np.array([self._get_value_for_pixel(px) for px in pxs], dtype=self.dtype)

    def _key_from_tile_filename(self, filename):
        """Get (x_grid, y_grid) key of upper left corner of tile from filename.

        Parameters
        ----------
        filename : str
            Tile filename. We assume filename is in format given by argument tile_structure,
            which is by default "%d_%d.tif" :
            'arbitrary_path/{x_grid}_{y_grid}.tif'
            e.g. 'data/tiled/2500_2250.tif'
            Path does not matter, but format after last slash is assumed.

        Returns
        -------
        x_grid : int
            minimum value for x for tile in raster pixel coordinates.

        y_grid : int
            minimum value for y for tile in raster pixel coordinates.
        """
        r = self.tile_regex.search(filename)
        if (len(r.groups()) != 2):
            raise ValueError("Tile filenames must comply with the given tile_structure : " + self.tile_structure)
        x_grid = int(r.group(1))
        y_grid = int(r.group(2))
        return x_grid, y_grid

    def _small_pixel_query(self, shp, shp_px):

        grid = self.to_geometry_grid(*shp_px.bounds)
        areas = {}
        for i, b in grid.items():

            if b.Intersects(to_geometry(shp, proj=self.proj)):
                diff = b.Intersection(to_geometry(shp, proj=self.proj))
                areas[i] = diff.GetArea()

        index = areas.keys()
        total_area = sum(areas.values())

        if total_area > 0:
            weights = np.array([areas[k]/self.pixel_area for k in index])
        else:
            weights = np.zeros(len(index))

        values = self.get_values_for_pixels(np.array(index))
        return values, weights

    def query(self, vector_layer, ext_outline=False, ext_fill=True,
              int_outline=False, int_fill=False, scale_factor=4,
              missing_first=False, small_polygon_pixels=4):
        """
        Query the dataset with a set of shapes (in a VectorLayer). The
        vectors will be reprojected into the projection of the raster. Any
        shapes in the vector layer that are not within the bounds of the
        raster will return with values and weights as np.array([]).

        Parameters
        ----------
        vector_layer : VectorLayer
            Set of shapes in vector format, with ids attached to each.
        ext_outline: boolean (default False)
            Include the outline of the shape in the raster
        ext_fill: boolean (default True)
            Fill the shape in the raster
        int_outline: booelan (default False)
            Include the outline of the interior shapes
        int_fill: boolean (default False):
            Fill the interior shapes
        scale_factor: int (default 4)
            The amount to scale the shape in X, Y before downscaling. The
            higher this number, the more precise the estimate of the overlap.

        missing_first: boolean (default false)
            Where the missing values should be at the beginning or the
            end of the results.

        small_polygon_pixels: integer (default 4)
            Number of pixels for the intersection of the polygon with the
            raster to be considered "small".  This is a slow step that computes
            the exact intersection between the polygon and the raster in the
            cooridate space of the raster (not pixel space!).

        Yields
        ------

        RasterQueryResult.  This is 3 attributes: id, values, weights.  The
        values are the pixel values from the raster.  the weights are the fraction
        of the pixel that is occupied by the polgon.
        """

        if self.proj.ExportToProj4() != vector_layer.proj.ExportToProj4():
            # Transform all vector shapes into raster projection.
            vl = vector_layer.transform(self.proj)
        else:
            vl = vector_layer

        # Filter out all shapes outside the raster bounds
        bbox = self.bbox()
        vl = vl.within(bbox)

        ids_to_tiles = None
        tiles_to_ids = None

        # Removing this for now.  It needs more thought!
        # Optimization to minimize memory usage if the RasterDataset contains
        # an index.  This will sort by the upper left corners of all the shapes
        # and process one shape at a time.  It will remove the corresponding
        # entries in self.raster_bands once all references for shapes in a
        # particular tile have been removed.
        #if self.index is not None:
        #    res = {self._key_from_tile_filename(id): set(vl.intersects(f).ids)
        #           for id, f in self.index.items()}

        #    tiles_to_ids = {k: v for k, v in res.items() if len(v) > 0}
        #    ids_to_tiles = defaultdict(set)
        #    for tile, shp_ids in tiles_to_ids.items():
        #        for id in shp_ids:
        #            ids_to_tiles[id].add(tile)

        #    vl = vl.sort()

        missing = vector_layer.index.difference(vl.index)

        if missing_first:
            ids = missing.append(vl.ids)
        else:
            ids = vl.ids.append(missing)

        px_shps = dict(zip(vl.ids, self.to_pixels(vl)))

        for id in ids:
            shp = px_shps.get(id, None)

            if shp is None:
                yield RasterQueryResult(id, np.array([]), [], np.array([]))

            else:
                #Eagerly load tiles
                #if ids_to_tiles is not None:
                #    for key in list(ids_to_tiles[id]):
                #        if key not in self.raster_bands:
                #            filename = self.path + "%d_%d.tif" % key
                #            self.raster_bands[key] = RasterBand(filename)
                #        tiles_to_ids[key].remove(id)

                # Check for small polygon since rasterizing a polygon
                # doesn't work for small polygons
                if vl[id].GetArea() < small_polygon_pixels * self.pixel_area:
                    values, weights = self._small_pixel_query(vl[id], shp)
                    coordinates = + np.array([])
                else:
                    # Rasterize the shape, and find list of all points.
                    mask = rasterize(shp, ext_outline=ext_outline,
                                     ext_fill=ext_fill,
                                     int_outline=int_outline,
                                     int_fill=int_fill,
                                     scale_factor=scale_factor).T

                    minx, miny, maxx, maxy = shp.bounds
                    idx = np.argwhere(mask > 0)

                    if idx.shape[0] == 0:
                        weights = mask[[0]]
                    else:
                        weights = mask[idx[:, 0], idx[:, 1]]

                    coordinates = (idx + np.array([minx, miny])).astype(int)
                    values = self.get_values_for_pixels(coordinates)

                yield RasterQueryResult(id, coordinates, values, weights)

            #if tiles_to_ids is None:
            #   continue

            #Remove raster bands that are empty
            #empty = [k for k, v in tiles_to_ids.items() if len(v) == 0]
            #for e in empty:
            #    del self.raster_bands[e]
            #    del tiles_to_ids[e]


class TiledWebRaster(RasterDataset):
    """
    Raster representation for tiled data sets commonly used on the web
    (e.g OpenLayers, GoogleMaps, etc.).  These have been assumed to be
    produced using the gdal2tiles.py script (found in /scripts dir).

    Assumes that the tiles are projected in Popular Visualisation CRS / Mercator
    (EPSG:3785)

    Parameters
    ----------
    path: str
       The path to the tiled data
    zoom: int
       Zoom to use, typically a value from 6 to 15.

    tile_size: int
       n x n size of each tile in pixels, default is 256

    bands: list of ints
       The bands to use.  Default is [1, 2, 3]

    xy_tile_path: str
       The tile path structure, default="%s/%s.png"

    Attributes
    ----------
    path : str
        Path to raster data files.

    resoltution : float
        Number of meters in both x & y that each pixel represents.
        For example, at a zoom of 11, each pixel is approx 76 m x 76 m.

    Notes
    -----
    Since these datasets are typically png files, there will be a substantial
    performance hit due to the CPU overhead of decompression
    """
    def __init__(self, path, zoom, tile_size=256, bands=None,
                 xy_tile_path="%s/%s.png"):
        self.gm = globalmaptiles.GlobalMercator(tileSize=256)
        self.resolution = self.gm.Resolution(zoom)
        self.zoom = zoom
        self.bands = range(1, 4) if bands is None else bands
        self.tile_size = tile_size
        xsize = 2**zoom*tile_size
        ysize = xsize
        max_lat = 20037508.342789244  # 1/2 the curcumference of the earther in meters
        min_lon = -20037508.342789244
        geo_transform = (min_lon, self.resolution, 0., max_lat, 0., -self.resolution)
        proj = projection_from_epsg(3785)  # Sphereical Mercator
        grid_size = (2*max_lat/self.resolution, 2*max_lat/self.resolution)
        tile_structure = "%d/" % zoom + xy_tile_path
        super(TiledWebRaster, self).__init__(path, xsize, ysize, geo_transform, proj,
                                             grid_size=grid_size, tile_structure=tile_structure)

    def _to_pixels(self, lat, lon, alt=None):
        a, b = super(TiledWebRaster, self)._to_pixels(lat, lon, alt=alt)
        return a, self.ysize - b

    def _get_value_for_pixel(self, px):
        """Look up value for a pixel in raster space.

        Parameters
        ----------
        px : np.array
            Pixel coordinates for 1 point: [x_coord, y_coord]

        """
        x_grid, y_grid = self.gm.PixelsToTile(px[0], px[1])

        # Look up the value in the x,y offset in the grid tile we just found
        # or read, and return it.
        x_px = px[0] - x_grid*self.tile_size
        y_px = px[1] - y_grid*self.tile_size

        # If we haven't already read this grid tile into memory, do so now,
        # and store it in raster_bands for future queries to access.
        if (x_grid, y_grid) not in self.raster_bands:
            filename = self.path + self.tile_structure % (x_grid, y_grid)
            self.raster_bands[(x_grid, y_grid)] = [read_vsimem(filename, b) for b in self.bands]
            if self.dtype is None:
                self.dtype = [r.dtype for r in self.raster_bands[(x_grid, y_grid)]]

        # Look up the grid tile for this pixel.
        raster = self.raster_bands[(x_grid, y_grid)]
        try:
            return [r[y_px][x_px] for r in raster]
        except:
            # TODO: get a default nan value for each dtype
            return [0 for i in raster]

    def get_values_for_pixels(self, pxs):
        """Look up values for a list of pixels in raster space.

        Parameters
        ----------
        pxs : np.array
            Array of pixel coordinates. Each row is [x_coord, y_coord]
            for one point.

        Returns
        -------
        list of dtype
            List of values in raster at pixel coordinates specified in pxs.
            Type is determined by GDAL2NP_CONVERSION from RasterBand data
            type.
        """
        values = [self._get_value_for_pixel(px) for px in pxs]
        n = len(values)
        m = len(self.bands)
        x = [[values[i][j] for i in range(n)] for j in range(m)]
        return [np.array(x[i], dtype=self.dtype[i]) for i in range(m)]


def read_catalog(dataset_catalog_filename_or_handle, workdir=None):
    """Take a catalog file and create a raster dataset

    Parameters
    ----------
    dataset_catalog_filename_or_handle : str or opened file handle
        if str : Path to catalog file for the dataset. May be relative or absolute.

        Catalog files are in json format, and usually represent a type of data
        (e.g. CDL) and a year (e.g. 2014).

    Returns
    -------
    RasterDataset

    See Also
    --------
    scripts/create_catalog.py : How to create a catalog file for a dataset.
    raster_query_test.py : Simple examples of exercising RasterQuery on tiled
        and untiled datasets, and computing stats from results.
    vector.py : Details of VectorLayer."""
    if isinstance(dataset_catalog_filename_or_handle, string_types):
        with open(dataset_catalog_filename_or_handle) as catalog_file:
            decoded = json.load(catalog_file)
    else:
        decoded = json.load(dataset_catalog_filename_or_handle)

    size = [int(x) for x in decoded["Size"]]
    coordinate_system = str(decoded["CoordinateSystem"])
    transform = decoded["GeoTransform"]

    # Get the projection for the raster
    proj = osr.SpatialReference()
    proj.ImportFromWkt(coordinate_system)

    if workdir is None:
        path = decoded["Path"]
    else:
        path = os.path.join(workdir, decoded["Path"])

    grid_size = decoded.get("GridSize", None)
    index = None
    tile_structure = None

    if "Index" in decoded:
        index, index_df = read_geojson(json.dumps(decoded["Index"]),
                                       index="location")
        index = index.transform(proj)

    if "Tile_structure" in decoded:
        tile_structure = decoded["Tile_structure"]

    return RasterDataset(path, size[0], size[1],
                         geo_transform=transform,
                         proj=proj,
                         grid_size=grid_size,
                         index=index,
                         tile_structure=tile_structure)


def read_raster(path, band_number=1):
    """
    Create a raster dataset from a single raster file

    Parameters
    ----------
    path: string
        Path to the raster file.  Can be either local or s3/gs.

    band_number: int
        The band number to use

    Returns
    -------

    RasterDataset
    """

    path = fileutils.get_path(path)
    ds = gdal.Open(path, GA_ReadOnly)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize
    proj = SpatialReference()
    proj.ImportFromWkt(ds.GetProjection())
    geo_transform = ds.GetGeoTransform()
    return RasterDataset(ds, xsize, ysize, geo_transform, proj)


def read_band(path, band_number=1):
    """
    Read a single band from a raster into memory.

    Parameters
    ----------
    path: string
        Path to the raster file.  Can be either local or s3/gs.

    band_number: int
        The band number to use

    Returns
    -------

    RasterBand
    """

    path = fileutils.get_path(path)
    ds = gdal.Open(path, GA_ReadOnly)
    return RasterBand(ds, band_number=band_number)


def read_vsimem(path, band_number=1):
    """
    Read a single band into memory from a raster. This method
    does not support all raster formats, only those that are
    supported by /vsimem

    Parameters
    ----------
    path: string
        Path to the raster file.  Can be either local or s3/gs.

    band_number: int
        The band number to use

    Returns
    -------

    RasterBand
    """
    filename = str(uuid4())
    with fileutils.open(path) as inf:
        gdal.FileFromMemBuffer("/vsimem/%s" % filename,
                               inf.read())

    ds = gdal.Open("/vsimem/%s" % filename)
    gdal.Unlink("/vsimem/%s" % filename)
    return RasterBand(ds, band_number=band_number)



"""
Copyright (c) 2016, Granular, Inc. 
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met: 

  * Redistributions of source code must retain the above copyright notice, this list of conditions 
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
    following disclaimer in the documentation and/or other materials provided with the distribution. 
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products 
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os
#Scipy stack
import numpy as np

#Spatial
from pyspatial.raster import rasterize
from pyspatial.raster import RasterBand
from pyspatial.vector import read_layer


class TestRasterBand:
    @classmethod
    def setup_class(cls):

        base = os.path.abspath(os.path.dirname(__file__))
        f = lambda x: os.path.join(base, x)
        # Let's add some CLUs for Washington County, IL

        vl, _ = read_layer(f("data/vector/clu/clu_public_a_il189.shp"),
                           index="uniqueid")

        # For speed, let's just use one shape.
        one_shape_id = "3850036893056913813r"
        cls.vl_one_shape = vl[[one_shape_id]]

        # Create a RasterBand for the raster. Raster data is stored and read from there.
        cls.rb = RasterBand(f("data/raster/95000_45000.tif"))

    @classmethod
    def teardown_class(cls):
        pass

    def get_counts_for_shape(self, shp):
        mask = rasterize(shp, ext_outline=0, int_outline=1).T
        minx, miny, maxx, maxy = shp.bounds
        pts = (np.argwhere(mask > 0) + np.array([minx, miny])).astype(int)
        return np.bincount(self.rb[pts[:, 1], pts[:, 0]], minlength=256)

    # Test that RasterBand internally converts coordinates to raster space,
    # so that the caller shouldn't need to do the conversion.
    def test_raster_band_should_convert_coordinates_single_shape(self):
        # Read in raster projection (applies to all raster tiles).
        rast_proj = self.rb.proj

        # Transform all vector shapes into raster projection, and convert to
        # pixels.
        vl_transformed = self.vl_one_shape.transform(rast_proj)
        px_shps_transformed = self.rb.to_pixels(vl_transformed)

        # Also convert CLUs to pixels without transforming into raster
        # projection.
        px_shps_not_transformed = self.rb.to_pixels(self.vl_one_shape)

        # Compute counts for a single shape using vl_transformed for reference.
        shp = px_shps_transformed[0]
        counts_transformed = self.get_counts_for_shape(shp)

        # Now, compute counts for a single shape using vl (which has not been
        # transformed), and compare against vl_transformed computation.
        shp = px_shps_not_transformed[0]
        counts_not_transformed = self.get_counts_for_shape(shp)
        assert(np.array_equal(counts_transformed, counts_not_transformed))



"""
Copyright (c) 2016, Granular, Inc. 
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted 
provided that the following conditions are met: 

  * Redistributions of source code must retain the above copyright notice, this list of conditions 
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the 
    following disclaimer in the documentation and/or other materials provided with the distribution. 
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products 
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS 
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT, 
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import numpy as np
from shapely.geometry import LineString, MultiLineString
from shapely.geometry import Polygon, MultiPolygon
from osgeo import osr, ogr

PROJ_WGS84 = "+proj=longlat +ellps=WGS84 +datum=WGS84 +no_defs"
ALBERS_N_AMERICA = ("+proj=aea +lat_1=20 +lat_2=60 +lat_0=40 "
                    "+lon_0=-96 +x_0=0 +y_0=0 +ellps=GRS80 +datum=NAD83 "
                    "+units=m +no_defs")


def projection_from_string(proj_str=PROJ_WGS84):
    """Returns a projection from a Proj4 string.
    If no argument is provided, uses WGS84/EPSG:4326"""
    srs = osr.SpatialReference()
    srs.ImportFromProj4(proj_str)
    return srs


def projection_from_epsg(epsg=4326):
    """Returns a projection from an epsg integer.
    If no argument is provided, uses 4326."""
    srs = osr.SpatialReference()
    srs.ImportFromEPSG(epsg)
    return srs


def projection_from_wkt(wkt=osr.SRS_WKT_WGS84):
    """Returns a projection from well known text.
    If no argument is provided, uses 4326 equivalent"""
    srs = osr.SpatialReference()
    srs.ImportfromWkt(wkt)
    return srs


# Create a function to project between spatial coords
def get_projection(obj, layer_index=0):
    if isinstance(obj, ogr.DataSource):
        obj = obj.GetLayerByIndex(layer_index)

    if hasattr(obj, "GetSpatialRef"):
        srs = obj.GetSpatialRef()
    else:
        proj_wkt = obj.GetProjection()
        srs = osr.SpatialReference()
        srs.ImportFromWkt(proj_wkt)
    return srs


"""
The SVG related functions in this file were adapted from the shapely project
(https://github.com/Toblerity/Shapely) which have been permitted for use
under the BSD license.
"""


def _repr_svg_(shp):
    """SVG representation for iPython notebook"""
    svg_top = ('<svg xmlns="http://www.w3.org/2000/svg" '
               'xmlns:xlink="http://www.w3.org/1999/xlink" ')
    if shp.is_empty:
        return svg_top + '/>'
    else:
        # Establish SVG canvas that will fit all the data + small space
        xmin, ymin, xmax, ymax = shp.bounds
        if xmin == xmax and ymin == ymax:
            # This is a point; buffer using an arbitrary size
            xmin, ymin, xmax, ymax = shp.buffer(1).bounds
        else:
            # Expand bounds by a fraction of the data ranges
            expand = 0.04  # or 4%, same as R plots
            widest_part = max([xmax - xmin, ymax - ymin])
            expand_amount = widest_part * expand
            xmin -= expand_amount
            ymin -= expand_amount
            xmax += expand_amount
            ymax += expand_amount
        dx = xmax - xmin
        dy = ymax - ymin
        width = min([max([100., dx]), 300])
        height = min([max([100., dy]), 300])
        try:
            scale_factor = max([dx, dy]) / max([width, height])
        except ZeroDivisionError:
            scale_factor = 1.
        view_box = "{0} {1} {2} {3}".format(xmin, ymin, dx, dy)
        transform = "matrix(1,0,0,-1,0,{0})".format(ymax + ymin)

        svg_templ = svg_top + (
            'width="{1}" height="{2}" viewBox="{0}" '
            'preserveAspectRatio="# XXX: MinYMin meet">'
            '<g transform="{3}">'
        ).format(view_box, width, height, transform)
        return scale_factor, svg_templ + "{0}</g></svg>"


def svg_polygon(shp, scale_factor, fill):
    """Returns SVG path element for the Polygon geometry.
    Parameters
    ==========
    scale_factor : float
        Multiplication factor for the SVG stroke-width.  Default is 1.
    fill_color : str, optional
        Hex string for fill color. Default is to use "#66cc99" if
        geometry is valid, and "#ff3333" if invalid.
    """
    if shp.is_empty:
        return '<g />'

    fill = fill if shp.is_valid else "#ff3333"
    stroke = fill
    exterior_coords = [
        ["{0:.4f},{1:.4f}".format(*c) for c in shp.exterior.coords]]
    interior_coords = [
        ["{0:.4f},{1:.4f}".format(*c) for c in interior.coords]
        for interior in shp.interiors]

    path = " ".join([
        "M {0} L {1} z".format(coords[0], " L ".join(coords[1:]))
        for coords in exterior_coords + interior_coords])
    return (
        '<path fill-rule="evenodd" fill="{2}" stroke="{3}" '
        'stroke-width="{0}" opacity="0.8" d="{1}" />'
        ).format(2. * scale_factor, path, fill, stroke)


def svg_multipolygon(shp, scale_factor, fill):
    """Returns group of SVG path elements for the MultiPolygon geometry.
    Parameters
    ==========
    scale_factor : float
        Multiplication factor for the SVG stroke-width.  Default is 1.
    fill_color : str, optional
        Hex string for fill color. Default is to use "#66cc99" if
        geometry is valid, and "#ff3333" if invalid.
    """
    if shp.is_empty:
        return '<g />'

    fill = fill if shp.is_valid else "#ff3333"
    return '<g>' + \
        ''.join(svg_polygon(p, scale_factor, fill) for p in shp) + \
        '</g>'


def svg_line(shp, scale_factor, stroke):
    """Returns SVG polyline element for the LineString geometry.
    Parameters
    ==========
    scale_factor : float
        Multiplication factor for the SVG stroke-width.  Default is 1.
    stroke_color : str, optional
        Hex string for stroke color. Default is to use "#66cc99" if
        geometry is valid, and "#ff3333" if invalid.
    """
    if shp.is_empty:
        return '<g />'

    stroke = stroke if shp.is_valid else "#ff3333"
    pnt_format = " ".join(["%0.4f,%0.4f" % c for c in shp.coords])
    return ('<polyline fill="none" stroke="{2}" stroke-width="{1}" '
            'points="{0}" opacity="0.8" />') \
        .format(pnt_format, 2. * scale_factor, stroke)


def svg_multiline(shp, scale_factor, stroke):
    """Returns a group of SVG polyline elements for the LineString geometry.
    Parameters
    ==========
    scale_factor : float
        Multiplication factor for the SVG stroke-width.  Default is 1.
    stroke_color : str, optional
        Hex string for stroke color. Default is to use "#66cc99" if
        geometry is valid, and "#ff3333" if invalid.
    """
    if shp.is_empty:
        return '<g />'

    stroke = stroke if shp.is_valid else "#ff3333"
    return '<g>' + \
        ''.join(svg_line(p, scale_factor, stroke) for p in shp) + \
        '</g>'


def to_svg(shp, color="#1f78b4"):
    """Function that takes a shapely object and returns
    an svg of that object.  Currently only supports (Multi)LineString,
    and (Multi)Polygon."""
    scale_factor, svg_str = _repr_svg_(shp)

    svg_fns = {LineString: svg_line,
               MultiLineString: svg_multiline,
               Polygon: svg_polygon,
               MultiPolygon: svg_multipolygon}

    fn = svg_fns.get(type(shp), None)
    if fn is None:
        return np.NaN

    svg = fn(shp, scale_factor, color)
    return svg_str.format(svg)



"""
Copyright (c) 2016, Granular, Inc.
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""
from __future__ import print_function
import pandas as pd
import numpy as np
from datetime import datetime, date
import re

# Python3
try:
    basestring
except NameError:
    basestring = str

TYPE_MAP = {float: "number",
            int: "number",
            basestring: "text",
            bool: "bool",
            pd.Timestamp: "datetime",
            date: "datetime",
            datetime: "datetime"}


def get_sample_pt(s):
    pt = s.dropna()
    if pt.shape[0] > 0:
        return pt.iloc[0]
    else:
        return None


def get_type(s, type_map=TYPE_MAP):
    t = type(get_sample_pt(s))

    if t in type_map:
        return type_map[t]
    elif np.dtype(t) == np.object:
        return None
    else:
        np_type = re.findall("[A-z]+", str(np.dtype(t)))[0]
        ret = None
        if np_type in ["int", "float"]:
            ret = "number"
        elif np_type == "S":
            ret = "text"
        else:
            ret = np_type

        assert ret in ["text", "number", "bool", "datetime"], "Invalid type for value: %s" % t
        return ret


def to_dict(df, hidden=None, not_visible=None, labels=None, types=None):
    """Convert a DataFrame into a Dataset dictionary.  This can later
    be serialized to json using the 'dumps' method found in this module.
    Types are either inferred or taken from the 'type' parameter.
    If the types cannot be identified, those columns will not be included.
    Note, if all the values of a column are NaN, the column will also
    be removed.

    Parameters
    -----------
    hidden: list (default=None)
        Columns in df to mark as hidden

    not_visible: list (default=None)
        Columns in df to not mark as visible

    labels: dict
        Keys are the column names and values are the 'label' attribute

    types: dict
        Keys are the column names and values are the 'type' attribute

    Returns
    -------
    Python dictionary with the attributes 'data', 'schema', 'index'
    """
    if df is None:
        df = pd.DataFrame()

    s = []
    if len(set(df.columns)) != len(df.columns):
        raise ValueError("DataFrame columns not unique!")

    hidden = [] if hidden is None else hidden
    not_visible = [] if not_visible is None else not_visible

    cols = []

    if df.index.name is None:
        cols.append("index")
        index = "index"
        hidden.append("index")
    else:
        index = df.index.name

    df = df.reset_index()

    for k, v in df.iteritems():
        row = {"type": get_type(v), "field": k, "label": k,
               "visible": True, "hidden": False}

        if types is not None and k in types:
            row["type"] = types[k]

        if row["type"] is None:
            msg = "Unable to determine type: %s." % k
            msg += " Please specify in types"
            print("WARN: " + msg)
            continue

        # Convert NaT to NaN
        if row["type"] == "datetime":
            df[k] = df[k].map(lambda x: np.NaN if pd.isnull(x) else x)

        if labels is not None and k in labels:
            row["label"] = labels[k]

        if k in hidden:
            row["hidden"] = True

        if k in not_visible:
            row["visible"] = False

        s.append(row)
        cols.append(k)

    return {"schema": s, "index": index,
            "data": df[cols].to_dict(orient="records")}


def dumps(x, double_precision=6):
    return pd.io.json.dumps(x, double_precision=double_precision,
                            iso_dates=True)



from smart_open import smart_open, ParseUri

from pyspatial.py3 import urlparse
import boto
from boto import connect_s3
import os
import six


def parse_uri(uri):
    parsed_uri = urlparse(uri)

    if parsed_uri.scheme in ['file', 's3', '']:
        return ParseUri(uri)
    elif parsed_uri.scheme in ['gs']:
        tmp_uri = ParseUri(uri.replace('gs://', 's3://'))
        tmp_uri.scheme = 'gs'
        return tmp_uri
    else:
        raise NotImplementedError("unknown URI scheme %r in %r" % (parsed_uri.scheme, uri))


def get_path(path):

    prefix = ""
    if path.endswith(".gz") and ".tar.gz" not in path:
        prefix = "/vsigzip/"

    uri = parse_uri(path)

    if uri.scheme == "file":
        path = uri.uri_path if os.path.exists(uri.uri_path) else None

    elif uri.scheme == "s3":
        conn = connect_s3()
        bucket = conn.get_bucket(uri.bucket_id)
        key = bucket.lookup(uri.key_id)
        if prefix == "":
            prefix = "/"
        prefix += os.path.join(prefix, "vsicurl")
        path = key.generate_url(60*60) if key is not None else key

    elif uri.scheme == 'gs':
        storage_uri = boto.storage_uri(uri.bucket_id, uri.scheme)
        bucket = storage_uri.get_bucket(uri.bucket_id)
        key = bucket.lookup(uri.key_id)
        if prefix == "":
            prefix = "/"
        prefix += os.path.join(prefix, "vsicurl/")
        path = key.generate_url(60*60) if key is not None else key

    return prefix+path


def open(path, mode="rb", **kw):
    uri = urlparse(path)

    if uri.scheme in ['file', 's3', '']:
        return smart_open(path, mode=mode, **kw)
    elif uri.scheme in ['gs']:
        if mode in ('r', 'rb'):
            storage_uri = boto.storage_uri(uri.netloc, uri.scheme)
            bucket = storage_uri.get_bucket(uri.netloc)
            key = bucket.get_key(uri.path)
            if key is None:
                raise KeyError(uri.path)
            return GSOpenRead(key, **kw)
        elif mode in ('w', 'wb'):
            storage_uri = boto.storage_uri(uri.netloc + '/' + uri.path, uri.scheme)
            key = storage_uri.new_key()
            if key is None:
                raise KeyError(uri.path)
            return GSOpenWrite(key, **kw)
        else:
            raise NotImplementedError("file mode %s not supported for %r scheme", mode, uri.scheme)
    else:
        raise NotImplementedError("scheme %r is not supported", uri.scheme)


class GSOpenRead(object):
    """
    Implement streamed reader from GS.
    """
    def __init__(self, read_key):
        if not hasattr(read_key, "bucket") and not hasattr(read_key, "name") and not hasattr(read_key, "read") \
                and not hasattr(read_key, "close"):
            raise TypeError("can only process GS keys")
        self.read_key = read_key

    def read(self, size=None):
        """
        Read a specified number of bytes from the key.
        """
        if not size or size < 0:
            # For compatibility with standard Python, `read(negative)` = read the rest of the file.
            # Otherwise, boto would read *from the start* if given size=-1.
            size = 0
        return self.read_key.read(size)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.read_key.close()

    def __str__(self):
        return "%s<key: %s>" % (
            self.__class__.__name__, self.read_key
        )

class GSOpenWrite(object):
    """
    Context manager for writing into GS files.
    """
    def __init__(self, outkey):
        if not hasattr(outkey, "bucket") and not hasattr(outkey, "name"):
            raise TypeError("can only process GS keys")

        self.outkey = outkey

    def __str__(self):
        return "%s<key: %s>" % (
            self.__class__.__name__, self.outkey
            )

    def write(self, b):
        """
        Write the given bytes (binary string) into the GS file from constructor.
        """
        if isinstance(b, six.text_type):
            # not part of API: also accept unicode => encode it as utf8
            b = b.encode('utf8')

        if not isinstance(b, six.binary_type):
            raise TypeError("input must be a binary string")

        self.outkey.set_contents_from_string(b)

    def close(self):
        self.outkey.close()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()



"""
Copyright (c) 2016, Granular, Inc.
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

"""
Parts of this file were adapted from the geopandas project
(https://github.com/geopandas/geopandas) which have been permitted for use
under the BSD license.
"""

from pyspatial.py3 import urlparse

import requests
from pyspatial import fileutils
from six import string_types
from numpy import ndarray
import pandas as pd
from rtree import index
from osgeo.osr import CoordinateTransformation, SpatialReference
from osgeo import ogr
from shapely.geometry.base import BaseGeometry
from shapely.geometry import shape
from shapely import wkb
from shapely.geometry import box
from shapely import ops
from pyspatial import utils as ut
from pyspatial.spatiallib import to_utm
from pyspatial.io import get_ogr_datasource, write_shapefile


def to_shapely(feat, proj=None):
    if isinstance(feat, BaseGeometry):
        return feat
    elif isinstance(feat, ogr.Feature):
        return wkb.loads(feat.geometry().ExportToWkb())
    elif isinstance(feat, list) or isinstance(feat, VectorLayer):
        if isinstance(feat[0], ogr.Feature):
            return [wkb.loads(f.geometry().ExportToWkb()) for f in feat]
        elif isinstance(feat[0], ogr.Geometry):
            return [wkb.loads(f.ExportToWkb()) for f in feat]
    elif isinstance(feat, ogr.Geometry):
        other = feat
        if proj is not None:
            other = feat.Clone()
            proj = feat.GetSpatialReference()
            other.TransformTo(proj)
        return wkb.loads(other.ExportToWkb())
    else:
        raise ValueError("Unable to convert to shapely object")


def to_geometry(shp, copy=False, proj=None):
    """Convert shp to a ogr.Geometry.

    Parameters
    ----------
    shp: ogr.Geometry, ogr.Feature, or shapely.BaseGeometry
        The shape you want to convert

    copy: boolean (default=False)
        Return a copy of the shape instead of a reference

    proj: str or osr.SpatialReference (default=None)
        The projection of the shape to define (if the shape is
        not projection aware), or transform to (if projection aware).
        If a string is provided, it assumes that it is in PROJ4.

    Returns
    -------
    ogr.Geometry"""

    target_proj = None
    source_proj = None

    # Check shape type
    if isinstance(shp, ogr.Geometry):
        geom = shp

    elif isinstance(shp, ogr.Feature):
        geom = shp.geometry()

    elif isinstance(shp, BaseGeometry):
        geom = ogr.CreateGeometryFromWkb(wkb.dumps(shp))
    else:
        raise ValueError("Unable to convert to ogr.Geometry object")

    # Check projection
    if isinstance(proj, string_types):
        target_proj = SpatialReference()
        target_proj.ImportFromProj4(proj)

    elif isinstance(proj, SpatialReference):
        target_proj = proj

    elif proj is None:
        target_proj = geom.GetSpatialReference()
        if target_proj is None:
            raise ValueError("shp does not have a SpatialReference")
    else:
        raise ValueError("Unable to set projction.")

    # Return shapely
    if isinstance(shp, BaseGeometry):
        geom.AssignSpatialReference(proj)
        return geom

    if copy:
        geom = geom.Clone()

    if proj is not None:
        source_proj = geom.GetSpatialReference()
        if source_proj is None:
            raise ValueError("shp does not have a SpatialReference")
        ct = CoordinateTransformation(source_proj, target_proj)
        geom.Transform(ct)
        geom.AssignSpatialReference(target_proj)

    return geom


def bounding_box(envelope, proj):
    xmin, xmax, ymin, ymax = envelope
    ring = ogr.Geometry(ogr.wkbLinearRing)
    ring.AddPoint(xmin, ymin)
    ring.AddPoint(xmin, ymax)
    ring.AddPoint(xmax, ymax)
    ring.AddPoint(xmax, ymin)
    ring.AddPoint(xmin, ymin)
    poly = ogr.Geometry(ogr.wkbPolygon)
    poly.AddGeometry(ring)
    poly.AssignSpatialReference(proj)
    return poly


def to_feature(shp, fid, proj=None):
    feature_def = ogr.FeatureDefn()
    feature = ogr.Feature(feature_def)
    feature.SetGeometry(to_geometry(shp, proj=proj))
    feature.SetFID(fid)
    return feature

# Handle issues with mock for sphinx
try:
    OLD_PANDAS = issubclass(pd.Series, ndarray)
except TypeError:
    OLD_PANDAS = True

BOOL_PREDICATES = ["intersects", "contains", "within", "crosses", "touches",
                   "equals", "disjoint"]


def _convert_array_args(args):
    _is_type = (isinstance(args[0], BaseGeometry) or
                isinstance(args[0], ogr.Geometry) or
                isinstance(args[0], ogr.Feature))

    if len(args) == 1 and _is_type:
        args = ([args[0]],)

    return args


def set_theoretic_methods(function, shp1, shp2):
    fns = ["Intersection", "Difference", "SymDifference", "Union"]
    assert function in fns, "function must be one of %s" % fns
    fn = getattr(shp1, function)
    proj1 = shp1.GetSpatialReference()
    proj2 = shp2.GetSpatialReference()

    other = shp2.Clone()

    # Reproject shp2 onto shp1
    if proj1.ExportToProj4() != proj2.ExportToProj4():
        other.TransformTo(proj1)

    return fn(other)


class VectorLayer(pd.Series):
    """
    Parameters
    ----------
    geometries: org.Feature[], ogr.Geometry[], shapely.BaseGeometry[]

    proj: osr.SpatialReference
         The projection for the geometries.  Defaults to EPSG:4326.

    index: iterable
        The index to use for the shapes

    Attributes
    ----------

    _sindex: rtree.index.Index
        The spatial index. Initially None, but can be built with build_sindex()



    """
    _metadata = ['name', 'proj']

    def __new__(cls, *args, **kwargs):
        kwargs.pop('crs', None)
        if OLD_PANDAS:
            args = _convert_array_args(args)
            arr = pd.Series.__new__(cls, *args, **kwargs)
        else:
            arr = pd.Series.__new__(cls)
        if type(arr) is VectorLayer:
            return arr
        else:
            return arr.view(VectorLayer)

    def __init__(self, *args, **kwargs):

        proj = kwargs.pop("proj", None)

        if proj is None:
            proj = ut.projection_from_epsg()

        if isinstance(args[0], pd.Series):
            kwargs.pop("index", None)

        super(VectorLayer, self).__init__(*args, **kwargs)

        self.proj = proj
        self._sindex = None

    @property
    def _constructor(self):
        return VectorLayer

    def _wrapped_pandas_method(self, mtd, *args, **kwargs):
        """Wrap a generic pandas method to ensure it returns a VectorLayer"""
        val = getattr(super(VectorLayer, self), mtd)(*args, **kwargs)
        if type(val) == pd.Series:
            val.__class__ = VectorLayer
            val.proj = self.proj
            val._sindex = None
        return val

    def __getitem__(self, key):
        return self._wrapped_pandas_method('__getitem__', key)

    def sort_index(self, *args, **kwargs):
        return self._wrapped_pandas_method('sort_index', *args, **kwargs)

    def take(self, *args, **kwargs):
        return self._wrapped_pandas_method('take', *args, **kwargs)

    def select(self, *args, **kwargs):
        return self._wrapped_pandas_method('select', *args, **kwargs)

    def _make_ids(self, ids):
        return pd.Index(ids)

    def append(self, *args, **kwargs):
        other = args[0]
        if self.proj.ExportToProj4() != other.proj.ExportToProj4():
            args = (other.transform(self.proj),)
        return self._wrapped_pandas_method('append', *args, **kwargs)

    # TODO: Fix this hack
    # Just to avoid a big refactor right now
    @property
    def features(self):
        return self

    # TODO: Fix this hack
    # Just to avoid a big refactor right now
    @property
    def ids(self):
        return self.index

    # TODO: add inplace support
    def filter_by_id(self, ids):
        """Return a vector layer with only those shapes with
        id in ids.

        Parameters
        ----------
        ids: iterable
            The ids to filter on"""

        assert hasattr(ids, "__iter__"), "ids must be iterable"
        if not isinstance(ids, pd.Index):
            ids = self._make_ids(ids)

        geoms = [self[i].Clone() for i in ids]
        proj = SpatialReference()
        proj.ImportFromWkt(self.proj.ExportToWkt())
        [g.AssignSpatialReference(proj) for g in geoms]
        return VectorLayer(geoms, index=ids)

    def _get_index_intersection(self, shp):
        if self._sindex is None:
            self.build_sindex()

        xmin, xmax, ymin, ymax = shp.GetEnvelope()
        bounds = (xmin, ymin, xmax, ymax)
        if isinstance(shp, list):
            raise ValueError("Collections of shapes are not supported!")

        return shp, self._sindex.intersection(bounds, objects="raw")

    def intersects(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that intersect with shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.intersects

        """

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Intersect(shp)]

        ids = self._make_ids(ids)

        if index_only:
            return ids
        else:
            return self.filter_by_id(ids)

    def iintersects(self, shp):
        """Return an index with only those shapes in the
        vector layer that intersect with shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------

        pandas.Index
        """
        return self.intersects(shp, index_only=True)

    def contains(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that contain shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.contains

        """
        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Contains(shp)]
        ids = self._make_ids(ids)

        if index_only:
            return ids
        else:
            return self.filter_by_id(ids)

    def icontains(self, shp):
        """Return an index with only those shapes in the
        vector layer that contain  shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------

        pandas.Index
        """
        return self.contains(shp, index_only=True)

    def within(self, shp, index_only=False):
        """Return a vector layer with only those shapes in
        the vector layer that are within shp.

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.within"""

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Within(shp)]
        ids = self._make_ids(ids)

        if index_only:
            return ids
        else:
            return self.filter_by_id(ids)

    def iwithin(self, shp):
        """Return an index with only those shapes in the
        vector layer that are within shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------

        pandas.Index"""
        return self.within(shp, index_only=True)

    def crosses(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that cross shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.crosses"""

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Crosses(shp)]
        ids = self._make_ids(ids)

        if index_only:
            return ids
        else:
            return self.filter_by_id(ids)

    def icrosses(self, shp):
        """Return an index with only those shapes in the
        vector layer that crosses shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------

        pandas.Index"""
        return self.crosses(shp, index_only=True)

    def touches(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that touches shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.touches
        """

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Touches(shp)]
        ids = self._make_ids(ids)

        if index_only:
            return ids
        else:
            return self.filter_by_id(ids)

    def itouches(self, shp):
        """Return an index with only those shapes in the
        vector layer that touches shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------

        pandas.Index"""
        return self.touches(shp, index_only=True)

    def equals(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that are equal shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#binary-predicates

        """

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = [i for i in ids if self[i].Equals(shp)]
        ids = self._make_ids(ids)

        if index_only:
            return ids

        return self.filter_by_id(ids)

    def iequals(self, shp):
        """Return an index with only those shapes in the
        vector layer that equals shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------
        pandas.Index"""
        return self.equals(shp, index_only=True)

    def disjoint(self, shp, index_only=False):
        """Return a vector layer with only those shapes in the
        vector layer that are disjoint with shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        index_only: boolean (default False)
            Return the ids only (not a new vector layer)

        See Also
        --------
        http://toblerity.org/shapely/manual.html#object.disjoint
        """

        shp = to_geometry(shp, proj=self.proj, copy=True)
        _shp, ids = self._get_index_intersection(shp)
        ids = self.index.difference(self._make_ids(ids))

        if index_only:
            return ids

        return self.filter_by_id(ids)

    def idisjoint(self, shp):
        """Return an index with only those shapes in the
        vector layer that is disjoint to shp

        Parameters
        ----------
        shp: shapely.BaseGeometry, ogr.Geometry, or ogr.Feature
            The shape to test against

        Returns
        -------
        pandas.Index"""
        return self.disjoint(shp, index_only=True)

    def _set_theoretic_methods(self, method, shp, reverse=False):
        vl = self.intersects(shp)

        shp = to_geometry(shp, proj=self.proj, copy=True)

        if isinstance(shp, list):
            raise ValueError("Collections of shapes are not supported!")

        geoms = []
        for geom in vl:
            a, b = (shp, geom) if reverse else (geom, shp)
            geoms.append(set_theoretic_methods(method, a, b))

        return VectorLayer(geoms, proj=self.proj, index=vl.index)

    def intersection(self, shp):
        """
        Cut the shapes in the VectorLayer to match the intersection
        specified by shp.

        Parameters
        ----------
        shp: shapely geometry or ogr Feature/Geometry

        Returns
        -------
        VectorLayer interesected by shp
        """
        return self._set_theoretic_methods("Intersection", shp)

    def difference(self, shp, kind="left"):
        """
        Cut the shapes in the VectorLayer to match the difference
        specified by shp.

        Parameters
        ----------
        shp: shapely geometry or ogr Feature/Geometry

        kind: str
           Either "left", "right", or "symmetric".
           In the case of "left" take geom.Difference(shp).
           In the case of "right", take shp.Difference(geom).
           Where geom is each geometry in the VectorLayer

        Returns
        -------
        VectorLayer
        """
        if kind == "left":
            return self._set_theoretic_methods("Difference", shp)
        elif kind == "right":
            return self._set_theoretic_methods("Difference", shp, reverse=True)
        elif kind == "symmetric":
            return self._set_theoretic_methods("SymDifference", shp)
        else:
            raise ValueError("kind must be one of {left, right, symmetric}")

    def symmetric_difference(self, shp, kind="left"):
        """
        Cut the shapes in the VectorLayer to match the symmetric difference
        specified by shp.

        Parameters
        ----------
        shp: shapely geometry or ogr Feature/Geometry

        Returns
        -------
        VectorLayer
        """
        return self.difference(shp, kind="symmetric")

    def union(self, shp):
        """
        Cut the shapes in the VectorLayer to match the union
        specified by shp.

        Parameters
        ----------
        shp: shapely geometry or ogr Feature/Geometry

        Returns
        -------
        VectorLayer interesected by shp
        """
        return self._set_theoretic_methods("Union", shp)

    def unary_union(self):
        return ops.unary_union(self.to_shapely())

    def is_valid(self, index_only=False):
        """
        Get vector layer with valid shapes.
        """
        ids = [i for i in self.index if self[i].IsValid]
        if index_only:
            return ids

        return self.filter_by_id(ids)

    def is_invalid(self, index_only=False):
        """
        Get vector layer with invalid shapes.
        """
        ids = [i for i in self.index if not self[i].IsValid]
        if index_only:
            return ids

        return self.filter_by_id(ids)

    def is_empty(self, index_only=False):
        """
        Get vector layer with the empty shapes
        """
        ids = [i for i in self.index if not self[i].IsEmpty]
        if index_only:
            return ids

        return self.filter_by_id(ids)

    def is_ring(self, index_only=False):
        """
        Get vector layer with the ring shapes
        """
        ids = [i for i in self.index if not self[i].IsRing]
        if index_only:
            return ids

        return self.filter_by_id(ids)

    def transform(self, target_proj):
        ct = CoordinateTransformation(self.proj, target_proj)
        geoms = [g.Clone() for g in self]
        [g.Transform(ct) for g in geoms]
        return VectorLayer(geoms, proj=target_proj, index=self.index)

    def to_wgs84(self):
        """Transform the VectorLayer into WGS84"""
        proj = ut.projection_from_epsg()
        return self.transform(proj)

    def to_shapely(self, ids=None):
        if ids is None:
            return self.map(to_shapely)
        else:
            if hasattr(ids, "__iter__"):
                return self[ids].map(to_shapely)
            else:
                return to_shapely(self[ids])

    def to_geometry(self, ids=None, proj=None):
        if ids is None:
            s = [to_geometry(f, proj=proj, copy=True) for f in self.features]
            ids = self.index
        else:
            if hasattr(ids, "__iter__"):
                s = [to_geometry(self[i], proj=proj, copy=True) for i in ids]
            else:
                return to_geometry(self[ids], proj=proj, copy=True)

        return pd.Series(s, index=ids)

    def map(self, f, as_geometry=False):
        """Apply a function, f, over all the geometries.

        Returns
        -------
        pandas.Series(as_geometry=False) or VectorLayer(as_geometry=True)
        """
        data = list(map(f, self.features))
        if not as_geometry:
            return pd.Series(data, index=self.index)
        else:
            return VectorLayer(data, index=self.index, proj=self.proj)

    def areas(self, proj=None):
        """Compute the areas for each of the shapes in the vector
        layer.

        Parameters
        ----------
        proj: string or osr.SpatialReference (default=None)
            valid strings are 'albers' or 'utm'. If None, no
            transformation of coordinates.

        Returns
        -------
        pandas.Series


        Note
        ----
        'utm' should only be used for small polygons when centimeter
        level accuraccy is needed.  Othewise the area will
        be incorrect.  Similar issues can happen when polygons cross
        utm boundaries.
        """
        if proj is None:
            return self.map(lambda x: x.GetArea())

        if proj == 'utm':
            if self.proj.ExportToProj4().strip() != ut.PROJ_WGS84:
                vl = self.transform(ut.projection_from_string())
            else:
                vl = self

            shps = vl.to_shapely()
            areas = [ops.transform(to_utm, shp).area for shp in shps]
            s = pd.Series(areas, index=self.index)
            s.name = "area_sqr_m"
            return s

        elif proj == 'albers':
            proj = ut.projection_from_string(ut.ALBERS_N_AMERICA)

        return self.transform(proj).areas()

    def distances(self, shp, proj=None):
        """Compute the euclidean distances for each of the shapes in the vector
        layer. If proj is not none, it will transform shp into proj.

        Note: if shp is a shapely object, it is upto to the user
        to make sure shp is in the correct coordinate system.

        Parameters
        ----------
        proj: string or osr.SpatialReference (default=None)
            valid strings are 'albers' or 'utm'. If None, no
            transformation of coordinates.

        Returns
        -------
        pandas.Series


        Note
        ----
        'utm' should only be used for small polygons when centimeter
        level accuraccy is needed.  Othewise the area will
        be incorrect.  Similar issues can happen when polygons cross
        utm boundaries.
        """
        if proj is None:
            shp = to_geometry(shp)
            return self.to_geometry(proj=proj).map(lambda x: x.Distance(shp))

        if proj == 'utm':
            if not self.proj.ExportToProj4() == ut.PROJ_WGS84:
                vl = self.transform(ut.projection_from_string())
            else:
                vl = self

            _shp = ops.transform(to_utm, to_shapely(shp))
            d = vl.to_shapely() \
                  .map(lambda x: ops.transform(to_utm, x).distance(_shp))
            s = pd.Series(d, index=self.index)
            return s

        elif proj == 'albers':
            proj = ut.projection_from_string(ut.ALBERS_N_AMERICA)

        shp = to_geometry(shp, copy=True, proj=proj)
        return self.to_geometry(proj=proj).map(lambda x: x.Distance(shp))

    def centroids(self, format="VectorLayer"):
        """Get a DataFrame with "x" and "y" columns for the
        centroid of each feature.

        Parameters
        ----------
        format: str (default='VectorLayer')
             Return type of the centroids.  available options are
             'Series', 'DataFrame', or 'VectorLayer'. 'Series'
             will a collection of (x, y) tuples.  'DataFrame' will
             be a DataFrame with columns 'x' and 'y'
        """

        formats = ["DataFrame", "VectorLayer", "Series"]

        if format in ("DataFrame", "Series"):
            data = (f.Centroid().GetPoints()[0] for f in self.features)
            if format == "Series":
                return pd.Series(data, index=self.index)
            else:
                return pd.DataFrame(data, columns=["x", "y"], index=self.index)
        elif format == "VectorLayer":
            pts = [g.Centroid() for g in self]
            [p.AssignSpatialReference(self.proj) for p in pts]
            return VectorLayer(pts, index=self.index, proj=self.proj)
        else:
            raise ValueError("format must be in %s" % formats)

    def envelopes(self):
        """The the envelope of each shape as xmin, xmax, ymin, ymax.
        Returns a pandas.Series."""
        data = (f.GetEnvelope() for f in self)
        return pd.Series(data, index=self.index)

    def boundingboxes(self):
        """Return a VectorLayer with the bounding boxes of each
        geometry"""
        geoms = self.envelopes().map(lambda x: bounding_box(x, self.proj))
        return VectorLayer(geoms, proj=self.proj, index=geoms.index)

    def upper_left_corners(self):
        """Get a DataFrame with "x" and "y" columns for the
        min_lon, max_lat of each feature"""
        data = [(f.GetEnvelope()[0], f.GetEnvelope()[3])
                for f in self.features]
        return pd.DataFrame(data, columns=["x", "y"], index=self.index)

    def size_bytes(self):
        """Get the size of the geometry in bytes"""
        return self.map(lambda x: x.WkbSize())

    def get_extent(self):
        """The xmin, xmax, ymin, ymax values of the layer"""
        if self._sindex is None:
            self.build_sindex()
        xmin, ymin, xmax, ymax = self._sindex.get_bounds()
        return (xmin, xmax, ymin, ymax)

    def bbox(self):
        """Return a geometry representing the bounding box of the layer"""
        (xmin, xmax, ymin, ymax) = self.get_extent()
        return to_geometry(box(xmin, ymin, xmax, ymax), proj=self.proj)

    def _gen_index(self):
        ix = range(len(self.features))
        for i, id, geom in zip(ix, self.index, self.features):
            xmin, xmax, ymin, ymax = geom.GetEnvelope()
            yield (i, (xmin, ymin, xmax, ymax), id)

    def items(self):
        return self.iteritems()

    def build_sindex(self):
        if self._sindex is None:
            self._sindex = index.Index(self._gen_index())

    def nearest(self, shp, max_neighbors=5):
        if isinstance(shp, BaseGeometry):
            xmin, ymin, xmax, ymax = shp.bounds
        elif isinstance(shp, ogr.Geometry):
            xmin, xmax, ymin, ymax = shp.GetEnvelope()
        elif isinstance(shp, ogr.Feature):
            xmin, xmax, ymin, ymax = shp.geometry().GetEnvelope()
        else:
            raise ValueError("Unable to compute bounds")

        self.build_sindex()

        neighbors = self._sindex.nearest((xmin, ymin, xmax, ymax),
                                         max_neighbors,
                                         objects="raw")
        ret = []
        i = 0
        while i < max_neighbors:
            try:
                ret.append(neighbors.next())
            except StopIteration:
                i = max_neighbors
            i += 1

        return ret

    def sort(self, kind="upper_left_corners", columns=["y", "x"],
             ascending=True, index_only=False):
        """Sort the vector layer by upper_left_corners or centroids

        Parameters
        ----------
        kind: str
            Either "upper_left_corners" or "centroids"

        columns : list of str (default ["y", "x"]
            Order in which to sort the shapes (ie. sort primarily by y-axis
            or x-axis).  Will be passed to pandas.DataFrame.sort.

        ascending : boolean (default True)
            Sort by columns in ascending or descending order.


        Returns
        -------
        list
            Shape ids sorted by columns.
        """

        kinds = set(["upper_left_corners", "centroids"])
        col_msg = "Sort columns must be in ['x', 'y']"
        assert all([c in ["x", "y"] for c in columns]), col_msg
        assert kind in kinds, "Sort kind not in %s" % ",".join(kinds)

        # Get dataframe with min_lon, max_lat for all shapes.
        df = getattr(self, kind)()
        df.sort_values(by=columns, ascending=ascending, inplace=True)

        if index_only:
            return df.index

        return self.filter_by_id(df.index)

    def to_dict(self, df=None):
        """Return a dictionary representation of the object.
        Based off the GeoJSON spec.  Will transform the vector
        layer into WGS84 (EPSG:4326).

        Parameters
        ----------
        df: pandas.DataFrame (default=None)
            The dataframe to supply the properties of the features.
            The index of df must match the ids of the VectorLayer.

        Returns
        -------
        dict
        """

        if self.proj.ExportToProj4() != ut.projection_from_string():
            vl = self.transform(ut.projection_from_string())
        else:
            vl = self

        res = {"type": "FeatureCollection"}
        res["features"] = [to_feature(f, i).ExportToJson(as_object=True)
                           for i, f in enumerate(vl)]

        if df is not None:
            for i, f in zip(vl.ids, res["features"]):
                props = f["properties"]
                df_props = df.loc[i].to_dict()
                f["properties"] = dict(props.items() + df_props.items())
                f["properties"]["__id__"] = i
        else:
            for i, f in zip(vl.ids, res["features"]):
                f["properties"]["__id__"] = i

        return res

    def to_json(self, path=None, df=None, precision=6):
        """Return the layer as a GeoJSON.  If a path is provided,
        it will save to the path. Otherwise, will return a string.

        Parameters
        ----------
        path: str, (default=None)
            The path to save the geojson data.
        df: pandas.DataFrame (default=None)
            The dataframe to supply the properties of the features.
            The index of df must match the ids of the VectorLayer.

        precision: int
            Number of decimal places to keep for floats

        Returns
        -------
        geojson string
        """
        res = self.to_dict(df=df)
        s = pd.io.json.dumps(res, double_precision=precision)
        if path is None:
            return s
        else:
            with fileutils.open(path, 'wb') as outf:
                outf.write(s)

    def to_svg(self, ids=None, ipython=False):
        """Return svg represention. Can output in IPython
        friendly form. If ids is None, will return a pandas.Series
        with all shapes as svg.

        Parameters
        ----------
        ids: str or iterable (default=None)
            The values of the geometries to convert to svg.
            If string, returns a string, if iterable, returns a Series.

        ipython: (default=False)
           Render in IPython friendly format

        Returns
        -------
        str or pandas.Series

        """

        if ipython:
            from IPython.display import HTML

        if ids is None:
            ids = self.index

        if hasattr(ids, "__iter__"):
            s = self.to_shapely(ids).map(ut.to_svg)
            if ipython:
                s = HTML("<br>".join(s))
            return s
        else:
            s = ut.to_svg(to_shapely(self[ids]))

        if ipython:
            return HTML(s)

        return s

    def to_shapefile(self, path, df=None):
        """Write the VectorLayer to an ESRI Shapefile.

        Only supports simple types for the attributes (int, float, str).  Any
        columns in the df that are not simple types will be ignored.

        Parameters
        ----------
        path: str
             Path to where you want to save the file. Can be local or s3/gs.

        df: pandas.DataFrame (default=None)
            Attach the attributes to the vector layer.  Similar to to_dict.
        """
        return write_shapefile(self, path, df=df)


def fetch_geojson(path):
    url = urlparse(path)
    if "http" in url.scheme:
        geojson = requests.get(path).text
    elif "s3" in url.scheme or url.scheme == "" or url.scheme == "file" or url.scheme == "gs":
        with fileutils.open(path) as inf:
            geojson = inf.read()
    else:
        return path
    return geojson


def read_datasource(ds, layer=0, index=None):
    dslayer = ds.GetLayerByIndex(layer)

    features = [dslayer.GetFeature(i) for i in
                range(dslayer.GetFeatureCount())]

    if index is None:
        ids = pd.Index([f.GetFID() for f in features])
    elif isinstance(index, str) or isinstance(index, unicode):
        ids = pd.Index([f[index] for f in features])
    elif hasattr(index, "__iter__"):
        ids = pd.Index(index)
    else:
        raise ValueError("Unable to create index.")

    if len(ids) != len(features):
        msg = "index length doesn't match number of shapes: %d vs %d."
        raise ValueError(msg % (len(ids), len(features)))

    rows = [f.items() for f in features]
    df = pd.DataFrame(rows, index=ids)
    geoms = [to_geometry(f, copy=True) for f in features]
    proj = ut.get_projection(dslayer)
    ds = None
    return VectorLayer(geoms, proj=proj, index=ids), df


def read_layer(path, layer=0, index=None):
    """Create a vector layer from the specified path.
    Will try to read using ogr.OpenShared.

    Parameters
    ----------
    path_or_str: string
        path or json string

    layer: integer
        The layer number to use.  Use ogrinfo to see
        the available layers.

    index: string or iterable (default=None)
        If string, the column in the "properties" of each feature to use
        as the index. If iterable, use the iterable as the index. If not
        specified, will create an integer based index.

    Returns
    -------
    Tuple of (VectorLayer, pandas.DataFrame of properties)
    """
    ds = get_ogr_datasource(path)
    return read_datasource(ds, layer=layer, index=index)


def read_geojson(path_or_str, index=None):
    """Create a vector layer from a geojson object.  Assumes that
    the data has a projection of EPSG:4326

    Parameters
    ----------
    path_or_str: string
        path or json string

    index: string or iterable
        If string, the column in the "properties" of each feature to use
        as the index. If iterable, use the iterable as the index.


    Returns
    -------

    Tuple of (VectorLayer, pandas.DataFrame of properties)"""

    if "FeatureCollection" not in path_or_str:
        geojson_str = fetch_geojson(path_or_str)
    else:
        geojson_str = path_or_str

    feats = pd.io.json.loads(geojson_str)["features"]

    if index is None:
        try:
            ids = [x["id"] for x in feats]
        except KeyError:
            ids = range(len(feats))

        name = "index"
    elif isinstance(index, string_types):
        ids = [x["properties"][index] for x in feats]
        name = index
    else:
        raise ValueError("Unable to create index.")

    proj = ut.projection_from_epsg()
    props = pd.DataFrame([x["properties"] for x in feats], index=ids)
    geoms = pd.Series([shape(x["geometry"]) for x in feats], index=ids) \
              .map(lambda x: to_geometry(x, proj=proj))

    props.index.name = name
    geoms.index.name = name

    return VectorLayer(geoms, proj=proj, index=ids), props


def from_series(geom_series, proj=None):
    """Create a VectorLayer from a pandas.Series object.  If
    the geometries do not have an spatial reference, EPSG:4326
    is assumed.

    Parameters
    ----------
    geom_series: pandas.Series
        The series object with shapely geometries

    proj: osr.SpatialReference
        The projection to use, defaults to EPSG:4326


    Returns
    -------
    VectorLayer

    """

    proj = ut.projection_from_string() if proj is None else proj
    geoms = geom_series.map(lambda x: to_geometry(x, proj=proj))
    return VectorLayer(geoms, proj=proj)



"""
Copyright (c) 2016, Granular, Inc.
All rights reserved.
License: BSD 3-Clause ("BSD New" or "BSD Simplified")

Redistribution and use in source and binary forms, with or without modification, are permitted
provided that the following conditions are met:

  * Redistributions of source code must retain the above copyright notice, this list of conditions
    and the following disclaimer.
  * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
    following disclaimer in the documentation and/or other materials provided with the distribution.
  * Neither the name of the nor the names of its contributors may be used to endorse or promote products
    derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import pandas as pd
from osgeo.ogr import Geometry
from shapely.geometry.base import BaseGeometry
import json

from jinja2 import Environment, PackageLoader
from pyspatial.dataset import to_dict, dumps
from pyspatial.utils import projection_from_epsg
from pyspatial import fileutils

STATIC_CSS = "https://granular-labs.s3.amazonaws.com/static/css"
STATIC_JS = "https://granular-labs.s3.amazonaws.com/static/js"
STATIC_IMG = "https://granular-labs.s3.amazonaws.com/static/img"


def to_feature(shp, _id):
    if isinstance(shp, BaseGeometry):
        return {"type": "Feature", "id": _id,
                "geometry": shp.__geo_interface__}

    elif isinstance(shp, Geometry):
        geom = shp.Clone()
        geom.TransformTo(projection_from_epsg())
        return {"type": "Feature", "id": _id,
                "geometry": json.loads(geom.ExportToJson())}
    elif pd.isnull(shp):
        return {"type": "Feature", "id": _id,
                "geometry": {"type": "Point",
                             "coordinates": []}}
    else:
        raise ValueError("Unable create feature dict from shp")


def get_geojson_dict(geo_data):
    features = None
    msg = "Unable to geo_data to convert to dict"

    if isinstance(geo_data, BaseGeometry) or isinstance(geo_data, Geometry):
        geo_data = [geo_data]

    elif isinstance(geo_data, str) or isinstance(geo_data, unicode):
        geo_data = json.loads(geo_data)

    if hasattr(geo_data, "__iter__") and len(geo_data) == 0:
        features = []

    # TODO: Find a better check with isinstance
    # since that would cause a circular reference
    # in vector.py
    elif hasattr(geo_data, "proj"):
        return geo_data.dropna().to_dict()

    elif isinstance(geo_data, pd.Series):
        g0 = geo_data[0]
        if isinstance(g0, Geometry) or isinstance(g0, BaseGeometry):
            features = [to_feature(s, i) for i, s in geo_data.items()]
        else:
            raise ValueError(msg)

    elif isinstance(geo_data, list):
        features = [to_feature(s, i) for i, s in enumerate(geo_data)]
    elif isinstance(geo_data, dict):
        if "type" not in geo_data and "features" not in geo_data:
            raise ValueError("Invalid geo_data")
        return geo_data

    else:
        raise ValueError(msg)

    return {"type": "FeatureCollection", "features": features}


def to_latlng(shp):
    if isinstance(shp, BaseGeometry):
        pt = to_feature(shp.centroid, 0)
    elif isinstance(shp, Geometry):
        pt = to_feature(shp.Centroid(), 0)
    elif pd.isnull(shp):
        return {"lat": None, "lng": None}
    else:
        raise ValueError("Unable to create latlng from shp")

    return dict(zip(["lng", "lat"], pt["geometry"]["coordinates"]))


def get_latlngs(geo_data):
    if isinstance(geo_data, BaseGeometry) or isinstance(geo_data, Geometry):
        return [to_latlng(geo_data)]

    elif hasattr(geo_data, "__iter__") and len(geo_data) == 0:
        return []

    elif hasattr(geo_data, "__iter__"):
        return [to_latlng(s) for i, s in geo_data.items()]

    else:
        raise ValueError("Invalid geo_data")


class HTMLMap(object):
    def __init__(self, lat, lng, zoom=5, data=None, info_cols=None,
                 height="100%", static_js=STATIC_JS, static_css=STATIC_CSS,
                 static_img=STATIC_IMG):
        self.lat = lat
        self.lng = lng
        self.zoom = zoom
        self.data = data
        self.info_cols = info_cols
        self.height = height
        self.base_layer = None
        self.choropleths = {}
        self.overlays = {}
        self.palettes = {}
        self.static_css = static_css
        self.static_js = static_js
        self.static_img = static_img

    def set_baselayer(self, geo_data):
        self.base_layer = get_geojson_dict(geo_data)

    def _render(self, height=None):
        if height is None:
            height = self.height

        env = Environment(loader=PackageLoader('pyspatial', 'templates'))
        data = {"base_layer": self.base_layer,
                "view": {"lat": self.lat, "lng": self.lng, "zoom": self.zoom},
                "overlays": self.overlays,
                "info_cols": self.info_cols,
                "dataset": to_dict(self.data),
                "choropleths": self.choropleths}

        data_json = dumps(data)
        opts = dict(DATA=data_json, height=height,
                    static_js=self.static_js, static_css=self.static_css,
                    static_img=self.static_img)

        self.html = env.get_template("map.html").render(**opts)

    def render_ipython(self, height="500px", width="100%"):
        from IPython.display import HTML
        self._render(height=height)
        """
        Embeds the HTML source of the report directly into an IPython notebook.
        """
        iframe = ('<iframe srcdoc="{srcdoc}" style="width: {width}; '
                  'height: {height}; border: none" id="some_id"></iframe>')
        srcdoc = self.html.replace('"', '&quot;')
        return HTML(iframe.format(width=width, height=height, srcdoc=srcdoc))

    def add_shapes(self, name, shapes, style=None, text=None):
        """Add shapes to the map. Accepts pandas Series or list of
        shapely or ogr.Geometry objects, or a dictionary or string
        matching the geojson spec."""
        if style is None:
            style = {}

        self.overlays[name] = {"shape": get_geojson_dict(shapes),
                               "style": style,
                               "kind": "geoJson",
                               "text": text}

    def add_markers(self, name, shapes, style=None, text=None):
        if style is None:
            style = {}

        self.overlays[name] = {"shape": get_latlngs(shapes),
                               "kind": "marker",
                               "style": style,
                               "text": text}

    def add_text(self, name, points, values, style=None):
        """Not implemented yet"""
        _style = {'background-color': 'none',
                  'font-size': '10pt',
                  'font-weight': 'bold',
                  'margin-top': '0'}

        if style is not None:
            _style.update(style)

        css = ";".join(["%s:%s" % (k, v) for k, v in _style.items()])
        style = "style=\"{css}\"".format(css=css)

    def choropleth(self, column=None, levels=6, palette="Reds"):
        '''
        Create a choropleth.

        #TODO:
        - Add style per choropleth.
        - Add support for different discretization schemes

        Parameters
        ----------
        column: str
            The column to use for the choropleth. Must exist in the data
            provided.

        levels: int, default None
            Number of levels to uses for the scale.  The allowed amount varies
            based on the palette that is chosen.
        palette: string or dict, default will use 'Reds'
             Uses color brewer (http://colorbrewer2.org/).
             If you pass a dict, the keys are the string values in the column,
             and the values are the html hex color for that value.
        '''

        if self.base_layer is None:
            raise ValueError("base_layer not set")

        if self.data is None:
            index = [f["id"] for f in self.base_layer["features"]]
            self.data = pd.DataFrame({"shapes": 1.}, index=index)
            levels = 3
            column = "shapes"

        if column not in self.data.columns:
            raise ValueError("%s not in %s" % (column, self.data.columns))

        if isinstance(palette, dict):
            self.choropleths[column] = palette
        else:
            self.choropleths[column] = {"palette": palette, "levels": levels}

    def save(self, path):
        """Save the html to a file.  Can be local or s3.  If s3, assumes that
        a .boto file exists"""
        with fileutils.open(path, 'wb') as outf:
            self._render()
            outf.write(self.html)
